package data;

import java.util.*;
import static helper.Global.*;
import static helper.Functions.*;
import structure.Structures.*;

/**
 * This class contains the method to generate the demand for CRS project. We generate two types of demand: for carshare
 * and rideshare, respectively, following the population distribution of the locations.
 *
 * @author  Miao Yu
 * @since September 14, 2017
 */
public class Demand {
    public static void main(String[] args){
        System.out.println("Program Start...");
        List<String[]> rawData = readDataCSV(baseDir+"census.csv");

        generateType1(rawData);
        //generateType2(rawData);

    }

    /*
    Generates Type 1 demand pool
     */
    private static void generateType1(List<String[]> rawData){
        System.out.println("Start generating Type 1 demand pool...");

        int size = 100;
        List<Type1> type1List = new ArrayList<Type1>(size);
        int i, j;


        // find potential service location
        int nSite = 5;
        List<Integer> sites = new ArrayList<>(5);
        // choose the sites based on the population of Type 2 users
        while(sites.size() != nSite) {
            int max = 0;
            int idx = -1;
            for (i = 0; i < rawData.size(); i++) {
                if (sites.contains(i)){
                    continue;
                }
                int pop = Integer.parseInt(rawData.get(i)[7]);
                if (pop > max) {
                    max = pop;
                    idx = i;
                }
            }
            sites.add(idx);
        }
        // calculate cumulative population;
        // calculate the demand based on the population of Type 1 user
        int[] cumulativePop = new int[sites.size()];
        for(i = 0 ; i < sites.size(); i++){
            for(j = 0; j <= i; j++){
                cumulativePop[i] += Integer.parseInt(rawData.get(sites.get(j))[5]);
            }
        }

        double[] distribution = {0.0, 0.1, 0.2, 0.3, 0.4};
        for (i = 0 ; i < size; i++){
            Type1 temp = new Type1();
            temp.origin= sites.get(pickFromProbability(cumulativePop));
            temp.destin = temp.origin;
            temp.start = random(9);
            //int duration = randomWithDistribution(distribution);
            int duration = random(3);
            temp.startD = temp.start + duration;
            temp.end = temp.start + duration + randomWithDistribution(distribution);// random(5);
            temp.startD = (temp.startD > 12)?12:temp.startD;
            temp.end = (temp.end > 12)?12:temp.end;
            if(temp.start==temp.end){
                i--;
            }else {
                type1List.add(temp);
            }
        }
        ArrayList<String> output = new ArrayList<>();
        output.add("Type 1 Demand"+" with 5 sites (Round-trip Only)");
        output.add("Based on Washtenaw County, Michigan");
        output.add("Count,"+size);
        output.add("origin, destination, startTime, startTimeDriveOther, endTime");
        for(Type1 e : type1List){
            int[] temp = {e.origin, e.destin, e.start, e.startD, e.end};
            output.add(combineStringCSV(temp));
        }
        writeDataCSV(baseDir,"stype1DemandPool", output);
    }


    /*
    Generate Type 2 demand pool
     */
    private static void generateType2(List<String[]> rawData){
        System.out.println("Start generating Type 2 demand pool...");
        int size = 100;
        List<Type2> type2List = new ArrayList<Type2>(size);
        int i, j;


        // find potential service location
        int[] population = new int[rawData.size()];
        for(i = 0; i < rawData.size(); i++){
            population[i] = Integer.parseInt(rawData.get(i)[7]);
        }

        int[] cumulativePop = cumulative(population);
        for (i = 0 ; i < size; i++){
            Type2 temp = new Type2();
            temp.origin= pickFromProbability(cumulativePop);
            temp.destin = pickFromProbability(cumulativePop);
            while(temp.origin == temp.destin){
                temp.destin = pickFromProbability(cumulativePop);
            }
            temp.tw_a = random(11*60);
            temp.tw_b = temp.tw_a+30;

            type2List.add(temp);
        }
        ArrayList<String> output = new ArrayList<>();
        output.add("Type 2 Demand");
        output.add("Based on Washtenaw County, Michigan");
        output.add("Count,"+size);
        output.add("origin, destination, openTimeWindow, closeTimeWindow");
        for(Type2 e : type2List){
            int[] temp = {e.origin, e.destin, e.tw_a, e.tw_b};
            output.add(combineStringCSV(temp));
        }
        writeDataCSV(baseDir,"type2DemandPool", output);

    }


}
