package data;

/**
 * This class contains the method to generate the traveling time matrix for Washtenaw community in Michigan. The traveling
 * time is calculated by accessing Google Map Distance Matrix API with valid API key. The output is stored in a csv file.
 *
 * @author Miao Yu
 * @since September 13, 2017
 */

import com.google.maps.*;
import com.google.maps.errors.ApiException;
import com.google.maps.model.DistanceMatrix;
import com.google.maps.model.LatLng;

import java.io.*;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

public class TravelMatrix {

    static List<Integer> pop = new ArrayList<Integer>();
    static List<Double> lat = new ArrayList<Double>();
    static List<Double> lon = new ArrayList<Double>();

    public static void main(String[] args) {
        LatLng temp;
        LatLng temp2;

        GeoApiContext gc = new GeoApiContext.Builder().apiKey("AIzaSyBjDQntHdFqxoYWjBtrHUCNeBpVxtc4fvY").build();

        readData();
        System.out.println(lat.size());
        System.out.println(lat.toString());
        System.out.println(lon.toString());
        int size = lat.size();
        long[][] travelMatrix = new long[size][size];
        for (int i = 0; i < size; i++) {
            for (int j = 0; j < size; j++) {
                if (i == j)
                    continue;
                else {
                    temp = new LatLng(lat.get(i), lon.get(i));
                    temp2 = new LatLng(lat.get(j), lon.get(j));
                    String[] origins = new String[1];
                    origins[0] = temp.toUrlValue();
                    String[] dests = new String[1];
                    dests[0] = temp2.toUrlValue();
                    try {
                        DistanceMatrix matrix = DistanceMatrixApi.getDistanceMatrix(gc, origins, dests).await();
                        travelMatrix[i][j] = matrix.rows[0].elements[0].duration.inSeconds;
                    } catch (InterruptedException e) {
                        e.printStackTrace();
                    } catch (IOException e) {
                        e.printStackTrace();
                    } catch (ApiException e) {
                        e.printStackTrace();
                    }
                }
            }
        }
        PrintWriter pw = null;
        try {
            pw = new PrintWriter(new File("data/base.csv"));
        } catch (FileNotFoundException e) {
            e.printStackTrace();
        }
        StringBuilder builder = new StringBuilder();
        builder.append("Washtenaw County\n");
        builder.append("size\n");
        builder.append(size);
        builder.append("\n");
        builder.append("Population\n");
        builder.append(pop.toString().replace("[", "").replace("]", ""));
        builder.append("\n");
        builder.append("Latitude\n");
        builder.append(lat.toString().replace("[", "").replace("]", ""));
        builder.append("\n");
        builder.append("Longitude\n");
        builder.append(lon.toString().replace("[", "").replace("]", ""));
        builder.append("\n");
        builder.append("Traveling Time Matrix\n");
        for (int i = 0; i < size; i++) {
            builder.append(Arrays.toString(travelMatrix[i]).replace("[", "").replace("]", ""));
            builder.append("\n");
        }
        pw.write(builder.toString());
        pw.close();
        System.out.println("done!");

    }



    public static void readData() {
        String csvFile = "data/data.csv";
        BufferedReader br = null;
        String line = "";
        String cvsSplitBy = ",";
        pop = new ArrayList<Integer>();
        lat = new ArrayList<Double>();
        lon = new ArrayList<Double>();
        try {

            br = new BufferedReader(new FileReader(csvFile));

            int counter = -1;

            while ((line = br.readLine()) != null) {
                if (counter == -1) {
                    counter++;
                    continue;
                }
                // use comma as separator
                String[] entry = line.split(cvsSplitBy);
                pop.add(Integer.parseInt(entry[2]));
                lat.add(Double.parseDouble(entry[8]));
                lon.add(Double.parseDouble(entry[9]));
                //System.out.println(Arrays.toString(entry));
                counter++;
            }

        } catch (FileNotFoundException e) {
            e.printStackTrace();
        } catch (IOException e) {
            e.printStackTrace();
        } finally {
            if (br != null) {
                try {
                    br.close();
                } catch (IOException e) {
                    e.printStackTrace();
                }
            }
        }
    }

}
