package solver;
import structure.Structures.Type1;
import structure.Structures.Type2;

import java.util.ArrayList;
import java.util.Arrays;

import static helper.Functions.readDataCSV;
import static helper.Global.*;
public class Input {
    // fileName
    public String fileName; // fileName
    public String dirName; // partial name
    public String fullName; // fileName with full directory

    public int nSite = 10; // number of service zones for carshare
    public int nType1; // number of Type 1 demand
    public int nType2; // number of Type 2 demand
    public int nVehicle;

    public double[][] travelMatrix; // base traveling matrix
    public ArrayList<Type1> type1List; // list of Type 1 demand
    public ArrayList<Type2> type2List; // list of Type 2 demand
    public double[][][] stochasticTravelMatrix;

    public int nScenarios;
    public int index;

    public void loadStochastic(int nScenarios){
        System.out.println("Start loading stochastic traveling matrix scenarios");
        System.out.println("total number to load: "+ nScenarios);
        stochasticTravelMatrix = new double[nScenarios][][];
        if(nScenarios == 1){
            ArrayList<String[]> rawData = readDataCSV(scenarioDir+"base"+".csv");
            int nLocation = Integer.parseInt(rawData.get(0)[1]);
            stochasticTravelMatrix[0] = new double[nLocation][nLocation];
            for(int i = 0; i < nLocation; i++){
                for(int j = 0; j < nLocation; j++){
                    stochasticTravelMatrix[0][i][j] = Double.parseDouble(rawData.get(i+1)[j]);
                }
            }
            return;
        }
        for(int k = 0; k < nScenarios; k++){
            ArrayList<String[]> rawData = readDataCSV(scenarioDir+k+".csv");
            int nLocation = Integer.parseInt(rawData.get(0)[1]);
            stochasticTravelMatrix[k] = new double[nLocation][nLocation];
            for(int i = 0; i < nLocation; i++){
                for(int j = 0; j < nLocation; j++){
                    stochasticTravelMatrix[k][i][j] = Double.parseDouble(rawData.get(i+1)[j]);
                }
            }
        }
    }

    public void loadSample(int nScenarios){
        System.out.println("Start loading stochastic traveling matrix scenarios");
        System.out.println("total number to load: "+ nScenarios);
        stochasticTravelMatrix = new double[nScenarios][][];

        for(int k = 0; k < nScenarios; k++){
            ArrayList<String[]> rawData = readDataCSV(sampleTestDir+k+".csv");
            int nLocation = Integer.parseInt(rawData.get(0)[1]);
            stochasticTravelMatrix[k] = new double[nLocation][nLocation];
            for(int i = 0; i < nLocation; i++){
                for(int j = 0; j < nLocation; j++){
                    stochasticTravelMatrix[k][i][j] = Double.parseDouble(rawData.get(i+1)[j]);
                }
            }
        }
    }

    public void loadStochastic(int s, int idx){
        this.nScenarios = s;
        this.index = idx;
        System.out.println("Start loading stochastic traveling matrix scenarios");
        System.out.println("total number to load: " + nScenarios);
        int position = index * 500;
        System.out.println("Start loading position: " + position);
        stochasticTravelMatrix = new double[nScenarios][][];
        for(int k = 0; k < nScenarios ; k++){
            ArrayList<String[]> rawData = readDataCSV(scenarioDir+(position+k)+".csv");
            int nLocation = Integer.parseInt(rawData.get(0)[1]);
            stochasticTravelMatrix[k] = new double[nLocation][nLocation];
            for(int i = 0; i < nLocation; i++){
                for(int j = 0; j < nLocation; j++){
                    stochasticTravelMatrix[k][i][j] = Double.parseDouble(rawData.get(i+1)[j]);
                }
            }
        }
    }

    public Input(String fullFileName){
        this.fullName = fullFileName;
        fullFileName = fullFileName.replace("\\", "/");
        String[] tokens = fullFileName.split("/");
        this.fileName = tokens[tokens.length - 1].substring(0, tokens[tokens.length-1].lastIndexOf('.'));
        this.dirName = tokens[tokens.length - 2]+ "/";
        System.out.println(fullName+" "+fileName);

        ArrayList<String[]> rawData = readDataCSV(fullName);
        System.out.println(rawData.get(0));
        nVehicle = Integer.parseInt(rawData.get(0)[1]);
        nType1 = Integer.parseInt(rawData.get(1)[1]);
        type1List = new ArrayList<>(nType1);
        for (int i = 0; i < nType1; i++){
            Type1 temp = new Type1(rawData.get(i+3));
            type1List.add(temp);
        }
        nType2 = Integer.parseInt(rawData.get(3+nType1)[1]);
        type2List = new ArrayList<>(nType2);
        for (int i = 0; i < nType2; i++){
            Type2 temp = new Type2(rawData.get(i+5+nType1));
            type2List.add(temp);
        }

        rawData = readDataCSV(scenarioDir+"base.csv");
        int nLocation = Integer.parseInt(rawData.get(0)[1]);
        travelMatrix = new double[nLocation][nLocation];
//        for(int i = 0; i < rawData.size(); i++)
//            System.out.println(Arrays.toString(rawData.get(i)));

        for(int i = 0; i < nLocation; i++){
            for(int j = 0; j < nLocation; j++){
                travelMatrix[i][j] = Double.parseDouble(rawData.get(i+1)[j]);
            }
        }
    }

    public static void main(String[] args){
        Input input = new Input("data/instance/c10-r10-k7/c10-r10-k7-n0.csv");
        System.out.println(input.type1List.size());
        System.out.println(input.type2List.size());
        System.out.println(Arrays.toString(input.travelMatrix[99]));
    }
}
