package solver;
import java.util.*;
import java.io.*;
import solver.Input;

import static helper.Global.*;
import static java.lang.Math.E;
import static helper.Functions.*;

public class Log {

    public Input input;
    public String fileName; // fileName
    public String dirName; // partial name
    public String fullName; // fileName with full directory

    public int nSite; // number of service zones for carshare
    public int nVehicle; // number of vehicle
    public int nType1; // number of Type 1 demand
    public int nType2; // number of Type 2 demand

    public double runtime_phaseI = 0;
    public double runtime_phaseII_deterministic = 0;
    public double runtime_phaseII_stochastic = 0;
    public double fulfillmentRate_Car = 0;
    public double fulfillmentRate_d_Ride = 0;
    public double fulfillmentRate_s_Ride = 0;
    //phase I solution
    public double[] rawSolution_x;
    public double[] rawSolution_y;
    public double[][] idleness;
    public double gap_phaseI;

    //phase II solution: deterministic
    public double[][] rawSolution_alpha;
    public double[] rawSolution_beta;
    public double[] rawSolution_gamma;
    public double gap_phaseII;
    public ArrayList<ArrayList<Integer>> assignment;
    public double[][][] relocation;

    //phase II solution: stochastic
    public double[][] rawSolution_s_alpha;
    public double[] rawSolution_s_beta;
    public double[] rawSolution_s_gamma;
    public double[][] rawSolution_wait;
    public double[][] rawSolution_idle;
    public double[][] rawSolution_overtime;
    public double gap_s_phaseII;

    public double totalWaitingTime[];
    public double totalIdleTime[];
    public double totalOverTime[];

    public int callback_count;
    public int subproblem_count;
    public double subproblem_time;

    public Log(Input ip){
        this.input = ip;
        this.fileName = ip.fileName;
        this.dirName = ip.dirName;
        this.fullName = resultDir+dirName+fileName;
        this.nSite = ip.nSite;
        this.nType1 = ip.nType1;
        this.nType2 = ip.nType2;
    }

    public void outputPhaseI(){
        ArrayList<String> output = new ArrayList<>();
        output.add("Instance Name, "+fileName);
        output.add("Raw Solution");
        output.add("x solution, (vehicle dispatch)");
        output.add(Arrays.toString(rawSolution_x).replace("[", "").replace("]", ""));
        output.add("y solution, (flow of vehicle)");
        output.add(Arrays.toString(rawSolution_y).replace("[", "").replace("]", ""));
        output.add("Type1 approval result");
        String temp = "";
        for(int l = 0; l < nType1; l++){
            temp += l+",";
        }
        output.add(temp);
        temp = "";
        for(int l = 0; l < nType1; l++){
            temp += rawSolution_y[l]+",";
        }
        output.add(temp+"");
        output.add("idle result");
        temp = "";
        output.add("Type 1, Time");
        for(int t = 0; t < 12; t++){
            temp += ","+t;
        }
        output.add(temp);
        for(int i = 0; i < nSite; i++){
            temp = i+","+Arrays.toString(idleness[i]).replace("[", "").replace("]", "");
            output.add(temp);
        }

        output.add("summary");
        output.add("nSite, nVehicle, nType1, nType2, Runtime, Type1ServiceRate" +
                ", Gap");

        output.add(input.nSite+","+input.nVehicle+","+input.nType1+","+input.nType2
                +","+runtime_phaseI+","+fulfillmentRate_Car+","+gap_phaseI+"");
        writeDataCSV(resultDir+dirName, fileName+"_phaseI", output);
        System.out.println("done!");
    }

    public void outputPhaseIRelocation(){
        ArrayList<String> output = new ArrayList<>();
        output.add("Instance Name, "+fileName);
        output.add("Raw Solution");
        output.add("x solution, (vehicle dispatch)");
        output.add(Arrays.toString(rawSolution_x).replace("[", "").replace("]", ""));
        output.add("y solution, (flow of vehicle)");
        output.add(Arrays.toString(rawSolution_y).replace("[", "").replace("]", ""));
        output.add("Type1 approval result");
        String temp = "";
        for(int l = 0; l < nType1; l++){
            temp += l+",";
        }
        output.add(temp);
        temp = "";
        for(int l = 0; l < nType1; l++){
            temp += rawSolution_y[l]+",";
        }
        output.add(temp+"");
        output.add("idle result");
        temp = "";
        output.add("Type 1, Time");
        for(int t = 0; t < 12; t++){
            temp += ","+t;
        }
        output.add(temp);
        for(int i = 0; i < nSite; i++){
            temp = i+","+Arrays.toString(idleness[i]).replace("[", "").replace("]", "");
            output.add(temp);
        }

        output.add("relocation");
        for(int t=0; t < 12; t++) {
            output.add("t="+t);
            for (int i = 0; i < nSite; i++) {
                output.add(Arrays.toString(relocation[t][i]).replace("[", "").replace("]", ""));
            }
        }

        output.add("summary");
        output.add("nSite, nVehicle, nType1, nType2, Runtime, Type1ServiceRate" +
                ", Gap");

        output.add(input.nSite+","+input.nVehicle+","+input.nType1+","+input.nType2
                +","+runtime_phaseI+","+fulfillmentRate_Car+","+gap_phaseI+"");
        writeDataCSV(resultDir+dirName, fileName+"_phaseI_relocation", output);
        System.out.println("done!");
    }





    public void outputPhaseII(){
        ArrayList<String> output = new ArrayList<>();
        output.add("Instance Name");
        output.add(input.fileName);
        output.add("raw alpha solution");
        for(int i = 0; i < rawSolution_alpha.length; i++){
            output.add(Arrays.toString(rawSolution_alpha[i]).replace("[", "").replace("]", ""));
        }
        output.add("raw beta solution");
        output.add(Arrays.toString(rawSolution_beta).replace("[", "").replace("]", ""));
        output.add("raw gamma solution");
        output.add(Arrays.toString(rawSolution_gamma).replace("[", "").replace("]", ""));
        output.add("Type 2 approval result");
        String temp = "";
        for(int l = 0; l < nType2; l++){
            temp += l+",";
        }
        output.add(temp.substring(0,temp.length()-1));
        output.add(Arrays.toString(rawSolution_beta).replace("[", "").replace("]", ""));
        output.add("assignment result");
        for(int l = 0; l < nType1; l++){
            temp = "";
            temp += l+",";
            for(int ele : assignment.get(l)){
                temp += ele+",";
            }
            output.add(temp.substring(0,temp.length()-1)+"\n");
        }
        output.add("scheduled time");
        double[] newArray = Arrays.copyOfRange(rawSolution_gamma, rawSolution_gamma.length - nType2, rawSolution_gamma.length);
        output.add(Arrays.toString(newArray).replace("[", "").replace("]", ""));
        output.add("summary");
        output.add("nSite, nVehicle, nType1, nType2, Runtime, Type2ServiceRate" +
                ", Gap");

        output.add(input.nSite+","+input.nVehicle+","+input.nType1+","+input.nType2
                +","+runtime_phaseII_deterministic+","+fulfillmentRate_d_Ride+","+gap_phaseII);
        writeDataCSV(resultDir+dirName, fileName+"_phaseII", output);
        System.out.println("done!");
    }

    public void outputPhaseIIStochastic(){
        ArrayList<String> output = new ArrayList<>();
        output.add("Instance Name");
        output.add(input.fileName);
        output.add("raw alpha solution");
        for(int i = 0; i < rawSolution_s_alpha.length; i++){
            output.add(Arrays.toString(rawSolution_s_alpha[i]).replace("[", "").replace("]", ""));
        }
        output.add("raw beta solution");
        output.add(Arrays.toString(rawSolution_s_beta).replace("[", "").replace("]", ""));
        output.add("raw gamma solution");
        output.add(Arrays.toString(rawSolution_s_gamma).replace("[", "").replace("]", ""));
        output.add("Type 2 approval result");
        String temp = "";
        for(int l = 0; l < nType2; l++){
            temp += l+",";
        }
        output.add(temp.substring(0,temp.length()-1));
        output.add(Arrays.toString(rawSolution_s_beta).replace("[", "").replace("]", ""));
        output.add("gamma solution");
        output.add(Arrays.toString(rawSolution_s_gamma).toString().replace("[", "").replace("]", ""));
        output.add("wait time solution");
        for(int i = 0; i < rawSolution_wait.length; i++){
            output.add(Arrays.toString(rawSolution_wait[i]).replace("[", "").replace("]", ""));
        }

        output.add("overtime solution");
        for(int i = 0; i < rawSolution_overtime.length; i++){
            output.add(Arrays.toString(rawSolution_overtime[i]).replace("[", "").replace("]", ""));
        }

        output.add("assignment result");
        for(int l = 0; l < nType1; l++){
            temp = "";
            temp += l+",";
            for(int ele : assignment.get(l)){
                temp += ele+",";
            }
            output.add(temp.substring(0,temp.length()-1));
        }
        output.add("scheduled time");
        double[] newArray = Arrays.copyOfRange(rawSolution_s_gamma, rawSolution_s_gamma.length-nType2, rawSolution_s_gamma.length);
        output.add(Arrays.toString(newArray).replace("[", "").replace("]", ""));
        output.add("summary");
        output.add("nSite, nVehicle, nType1, nType2, Runtime, Type2ServiceRate, Gap");


        output.add(input.nSite+","+input.nVehicle+","+input.nType1+","+input.nType2
                +","+runtime_phaseII_stochastic+","+fulfillmentRate_s_Ride+","+gap_s_phaseII);
        writeDataCSV(resultDir+dirName, fileName+"_Stochastic_"+input.nScenarios+"_"+input.index, output);
        System.out.println("done!");
    }

    public void outputPhaseIIBender(){
        ArrayList<String> output = new ArrayList<>();
        output.add("Instance Name");
        output.add(input.fileName);

        output.add("Training Scenarios");
        output.add(Integer.toString(input.nScenarios));
        output.add("Index");
        output.add(Integer.toString(input.index));
//        output.add("raw alpha solution");
////        for(int i = 0; i < rawSolution_s_alpha.length; i++){
////            output.add(Arrays.toString(rawSolution_s_alpha[i]).replace("[", "").replace("]", ""));
////        }
////        output.add("raw beta solution");
////        output.add(Arrays.toString(rawSolution_s_beta).replace("[", "").replace("]", ""));
////        output.add("raw gamma solution");
////        output.add(Arrays.toString(rawSolution_s_gamma).replace("[", "").replace("]", ""));
////        output.add("Type 2 approval result");
        String temp = "";
//        for(int l = 0; l < nType2; l++){
//            temp += l+",";
//        }
//        output.add(temp.substring(0,temp.length()-1));
//        output.add(Arrays.toString(rawSolution_s_beta).replace("[", "").replace("]", ""));
//        output.add("gamma solution");
//        output.add(Arrays.toString(rawSolution_s_gamma).toString().replace("[", "").replace("]", ""));
//        output.add("wait time solution");
//        for(int i = 0; i < rawSolution_wait.length; i++){
//            output.add(Arrays.toString(rawSolution_wait[i]).replace("[", "").replace("]", ""));
//        }
//
//        output.add("overtime solution");
//        for(int i = 0; i < rawSolution_overtime.length; i++){
//            output.add(Arrays.toString(rawSolution_overtime[i]).replace("[", "").replace("]", ""));
//        }

        output.add("assignment result");
        for(int l = 0; l < nType1; l++){
            temp = "";
            temp += l+",";
            for(int ele : assignment.get(l)){
                temp += ele+",";
            }
            output.add(temp.substring(0,temp.length()-1));
        }
        output.add("scheduled time");
        output.add(Arrays.toString(rawSolution_s_gamma).replace("[", "").replace("]", ""));
        output.add("summary");
        output.add("nSite, nVehicle, nType1, nType2, Runtime, Type2ServiceRate, Gap, callback_count, subproblem_count, subproblem_time");


        output.add(input.nSite+","+input.nVehicle+","+input.nType1+","+input.nType2
                +","+runtime_phaseII_stochastic+","+fulfillmentRate_s_Ride+","+gap_s_phaseII
                +","+callback_count+","+subproblem_count+","+subproblem_time);
        writeDataCSV(resultDir+dirName, fileName+"_Bender_"+input.nScenarios+"_"+input.index, output);
        System.out.println("done!");
    }

    public void outputPhaseIIBenders(){
        ArrayList<String> output = new ArrayList<>();
        output.add("Instance Name");
        output.add(input.fileName);

        output.add("Training Scenarios");
        output.add(Integer.toString(input.nScenarios));
        output.add("Index");
        output.add(Integer.toString(input.index));
//        output.add("raw alpha solution");
////        for(int i = 0; i < rawSolution_s_alpha.length; i++){
////            output.add(Arrays.toString(rawSolution_s_alpha[i]).replace("[", "").replace("]", ""));
////        }
////        output.add("raw beta solution");
////        output.add(Arrays.toString(rawSolution_s_beta).replace("[", "").replace("]", ""));
////        output.add("raw gamma solution");
////        output.add(Arrays.toString(rawSolution_s_gamma).replace("[", "").replace("]", ""));
////        output.add("Type 2 approval result");
        String temp = "";
//        for(int l = 0; l < nType2; l++){
//            temp += l+",";
//        }
//        output.add(temp.substring(0,temp.length()-1));
//        output.add(Arrays.toString(rawSolution_s_beta).replace("[", "").replace("]", ""));
//        output.add("gamma solution");
//        output.add(Arrays.toString(rawSolution_s_gamma).toString().replace("[", "").replace("]", ""));
//        output.add("wait time solution");
//        for(int i = 0; i < rawSolution_wait.length; i++){
//            output.add(Arrays.toString(rawSolution_wait[i]).replace("[", "").replace("]", ""));
//        }
//
//        output.add("overtime solution");
//        for(int i = 0; i < rawSolution_overtime.length; i++){
//            output.add(Arrays.toString(rawSolution_overtime[i]).replace("[", "").replace("]", ""));
//        }

        output.add("assignment result");
        for(int l = 0; l < nType1; l++){
            temp = "";
            temp += l+",";
            for(int ele : assignment.get(l)){
                temp += ele+",";
            }
            output.add(temp.substring(0,temp.length()-1));
        }
        output.add("scheduled time");
        output.add(Arrays.toString(rawSolution_s_gamma).replace("[", "").replace("]", ""));
        output.add("summary");
        output.add("nSite, nVehicle, nType1, nType2, Runtime, Type2ServiceRate, Gap, callback_count, subproblem_count, subproblem_time");


        output.add(input.nSite+","+input.nVehicle+","+input.nType1+","+input.nType2
                +","+runtime_phaseII_stochastic+","+fulfillmentRate_s_Ride+","+gap_s_phaseII
                +","+callback_count+","+subproblem_count+","+subproblem_time);
        writeDataCSV(resultDir+dirName, fileName+"_Benders_new_"+input.nScenarios+"_"+input.index, output);
        System.out.println("done!");
    }

}