package solver;


import gurobi.*;
import structure.Structures.Edge2;
import structure.Structures.Node;
import structure.Structures.Type1;
import structure.Structures.Type2;

import java.io.File;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import static helper.Global.logDir;
import static helper.Global.ratePerHourRide;

public class Phase2Bender2 {

    public int nScenario;
    public int index;
    public double lowerBound;
    public double upperBound;
    public double[][] bestAlphaSol;
    public double[] bestBetaSol;
    public double[] bestGammaSol;
    public class Timer {
        public int counter;
        public int counter_sub;
        public long time;
        public long overallTime;
        public Timer (){
            counter = 0;
            counter_sub = 0;
            time = 0;
            overallTime = 0;
        }
    }


    public Input input;
    public Log result;
    public File f;
//    public double ratePerHourCar = 8;
//    public double ratePerHourRide = 30;
//    public double costIdle = 0;
//    public double costService = 0;
//    public double rateDriveOther = 15;


    public Phase2Bender2(){

    }

    public List<Integer> vehicle;
    public List<Node> V0 = new ArrayList<Node>(); // depot start
    public List<Node> V1 = new ArrayList<Node>(); // depot end
    public List<Node> V2 = new ArrayList<Node>(); // ride
    public List<Edge2> E0 = new ArrayList<Edge2>(); // leave depot
    public List<Edge2> E1 = new ArrayList<Edge2>(); // travel
    public List<Edge2> E2 = new ArrayList<Edge2>(); // return depot
    public List<Node> V = new ArrayList<Node>();
    public List<Edge2> E = new ArrayList<Edge2>();

    public int nCar;
    public int nRide;

    /**
     * Constructor of stochastic phase II solver
     * @param i
     * @param nScenario
     * @param index
     */
    public Phase2Bender2(Input i, int nScenario, int index){

        this.input = i;
        this.nScenario = nScenario;
        this.index = index;
        this.result = new Log(input);
        this.vehicle = new ArrayList<Integer>();
        for(int l = 0; l < input.nType1; l++){
            if(input.type1List.get(l).approved){
                vehicle.add(l);
            }
        }
        //generate nodes
        V0 = new ArrayList<Node>(); // depot start
        V1 = new ArrayList<Node>(); // depot end
        V2 = new ArrayList<Node>(); // ride
//        List<Node> V3 = new ArrayList<Node>();
        nCar = input.nType1;
        nRide = input.nType2;
        int counter = 0;
        for(int l : vehicle){
            Type1 cr = input.type1List.get(l);
            Node node = new Node();
            node.index = counter;
            node.location = cr.origin;
            node.tw_a = cr.startD * 60;
            node.tw_b = cr.end * 60;
            node.idx_car = l;
            V0.add(node);
            counter++;
        }

        for(int l : vehicle){
            Type1 cr = input.type1List.get(l);
            Node node = new Node();
            node.index = counter;
            node.location = cr.destin;
            node.tw_a = cr.startD * 60;
            node.tw_b = cr.end * 60;
            node.idx_car = l;
            V1.add(node);
            counter++;
        }

        for(int j = 0; j < nRide; j++){
            Type2 rr = input.type2List.get(j);
            Node node = new Node();
            node.index = counter;
            node.location = rr.origin;
            node.tw_a = rr.tw_a;
            node.tw_b = rr.tw_b;
            node.idx_ride = j;
            V2.add(node);
            counter++;
        }

        // generate Edge
        E0 = new ArrayList<Edge2>(); // leave depot
        E1 = new ArrayList<Edge2>(); // travel
        E2 = new ArrayList<Edge2>(); // return depot
        //List<Edge2> E3 = new ArrayList<Edge2>();

        counter = 0;
        for(Node u : V0){
            for(Node v : V2){
                Edge2 edge = new Edge2();
                edge.node1 = u;
                edge.node2 = v;
                edge.index = counter;
                E0.add(edge);
                counter++;
            }
        }

        for(Node u : V2){
            for(Node v : V2){
                if(u.idx_ride == v.idx_ride){
                    continue;
                }else{
                    Edge2 edge = new Edge2();
                    edge.node1 = u;
                    edge.node2 = v;
                    edge.index = counter;
                    E1.add(edge);
                    counter++;
                }
            }
        }


        for(Node u : V2){
            for(Node v : V1){
                Edge2 edge = new Edge2();
                edge.node1 = u;
                edge.node2 = v;
                edge.index = counter;
                E2.add(edge);
                counter++;
            }
        }

        V = new ArrayList<Node>();
        V.addAll(V0);
        V.addAll(V1);
        V.addAll(V2);
        E = new ArrayList<Edge2>();
        E.addAll(E0);
        E.addAll(E1);
        E.addAll(E2);

        lowerBound = Double.MIN_VALUE;
        upperBound = Double.MAX_VALUE;
    }


    protected ArrayList<ArrayList<Integer>> findsubtour(double[][] alp) {
        ArrayList<ArrayList<Integer>> result = new ArrayList<>();
        int nNode = V.size();
        int nEdge = E.size();
        double[][] sol = new double[nNode][nNode];
        for(int l = 0; l < vehicle.size(); l++){
            for(Edge2 e : E){
                if(alp[e.index][l] > 0.9){
                    sol[e.node1.index][e.node2.index] += 1;
                }
            }
        }

        boolean[] visited  = new boolean[nNode];
        int[] pred = new int[nNode];
        for(int i = 0; i < nNode; i++)
            pred[i] = -1;
        for(int i = 0; i < nNode; i++){
            if(visited[i]){
                continue;
            }
            boolean flag = true;
            int cur = i;
            while(flag){
                for(int j = 0; i < nNode; i++){
                    flag = false;
                    if(sol[i][j] > 0.9){
                        if(visited[j] == false){
                            flag = true;
                            pred[j] = cur;
                            cur = j;

                        }
                    }
                }
            }
        }


        return result;
    }


    public ArrayList<ArrayList<Integer>> routes(double[][] sol){
        ArrayList<ArrayList<Integer>> result = new ArrayList<>();
        for(int k = 0; k < input.nVehicle; k++){
            result.add(new ArrayList<>());
            int cur = V0.get(k).index;
            result.get(k).add(cur);
            boolean flag = true;
            while(flag){
                flag = false;
                for(int i = V1.size(); i < V.size(); i++){
                    if(sol[cur][i] > 0.5){
                        result.get(k).add(i);
                        cur = i;
                        flag = true;
                    }
                }
            }
            result.get(k).add(V1.get(k).index);
        }

        return result;
    }


    public double subproblem(ArrayList<Integer> order, int vehicle, GRBEnv sub){
        try{

            GRBModel submodel = new GRBModel(sub);
            GRBVar[] gamma = new GRBVar[order.size()+2];
            gamma[0] = submodel.addVar(V0.get(vehicle).tw_a , V0.get(vehicle).tw_b, 0, GRB.CONTINUOUS, "schedule0");
            for(int i = 1; i < order.size()+1; i++){
                gamma[i] = submodel.addVar(V2.get(order.get(i-1)).tw_a , V2.get(order.get(i-1)).tw_b ,0, GRB.CONTINUOUS, "schedule"+i);
            }
            gamma[order.size()+1] = submodel.addVar(V1.get(vehicle).tw_a , V1.get(vehicle).tw_b , 0, GRB.CONTINUOUS, "schedule"+(order.size()+1));

            GRBVar[][] actual = new GRBVar[nScenario][order.size()+2];
            for(int s = 0; s < nScenario; s++){
                for(int i = 0; i < order.size()+2; i++){
                    actual[s][i] = submodel.addVar(0, GRB.INFINITY, 0., GRB.CONTINUOUS, "wait"+s+","+i);
                }
            }
            GRBVar[][] wait = new GRBVar[nScenario][order.size()+1];
            for(int s = 0; s < nScenario; s++){
                for(int i = 0; i < order.size()+1; i++){
                    wait[s][i] = submodel.addVar(0, GRB.INFINITY, 0., GRB.CONTINUOUS, "wait"+s+","+i);
                }
            }
            GRBVar[] overtime = new GRBVar[nScenario];
            for(int s = 0; s < nScenario; s++){
                overtime[s] = submodel.addVar(0,GRB.INFINITY, 0, GRB.CONTINUOUS, "over"+s);
            }

            GRBLinExpr expr = new GRBLinExpr();
            for(int s = 0; s < nScenario; s++){
                expr.addTerm(0.1/nScenario, overtime[s]);
                for(int j = 1; j < order.size()+1; j++){
                    expr.addTerm(0.1/nScenario, wait[s][j]);
                }
            }
            submodel.setObjective(expr, GRB.MINIMIZE);

            for(int s = 0; s < nScenario; s++){
                for(int i = 0; i < order.size()+1; i++){
                    expr.clear();
                    expr.addTerm(1, actual[s][i]);
                    expr.addTerm(-1, gamma[i]);
                    expr.addTerm(-1, wait[s][i]);
                    submodel.addConstr(expr,GRB.EQUAL, 0, "wait_bind"+s+","+i);
                }

                expr.clear();
                expr.addTerm(1, actual[s][order.size()+1]);
                expr.addTerm(-1, overtime[s]);
                submodel.addConstr(expr, GRB.EQUAL, V1.get(vehicle).tw_b, "overtime_bind"+s);

                for(int i = 0; i < order.size()+1; i++){
                    expr.clear();
                    expr.addTerm(1, actual[s][i+1]);
                    expr.addTerm(-1, actual[s][i]);

                    double duration;
                    double trip;
                    if(i == 0){
                        duration = 0;
                        Node u = V0.get(vehicle);
                        Node v = V2.get(order.get(i));
                        trip = input.stochasticTravelMatrix[s][u.location][v.location];
                    }else {
                        Type2 temp = input.type2List.get(order.get(i-1));
                        duration = input.stochasticTravelMatrix[s][temp.origin][temp.destin];
                        Node u = V2.get(order.get(i-1));
                        Node v = i >= order.size()? V1.get(vehicle):V2.get(order.get(i));
                        trip = input.stochasticTravelMatrix[s][u.location][v.location];
                    }
                    submodel.addConstr(expr, GRB.GREATER_EQUAL, trip + duration, "connect"+s+i );
                }
            }

            submodel.optimize();
            double obj = submodel.get(GRB.DoubleAttr.ObjVal);

            submodel.dispose();
            return obj;

        }catch (GRBException e) {
            System.out.println("Error code: " + e.getErrorCode() + ". " +
                    e.getMessage());
        }

        return 0;
    }

    public ArrayList<Double> subproblem2(ArrayList<Integer> order, int vehicle){
        try{
            GRBEnv env = new GRBEnv();
            GRBModel submodel = new GRBModel(env);
            GRBVar[] gamma = new GRBVar[order.size()+2];
            gamma[0] = submodel.addVar(V0.get(vehicle).tw_a , V0.get(vehicle).tw_b, 0, GRB.CONTINUOUS, "schedule0");
            for(int i = 1; i < order.size()+1; i++){
                gamma[i] = submodel.addVar(V2.get(order.get(i-1)).tw_a , V2.get(order.get(i-1)).tw_b ,0, GRB.CONTINUOUS, "schedule"+i);
            }
            gamma[order.size()+1] = submodel.addVar(V1.get(vehicle).tw_a , V1.get(vehicle).tw_b , 0, GRB.CONTINUOUS, "schedule"+(order.size()+1));

            GRBVar[][] actual = new GRBVar[nScenario][order.size()+2];
            for(int s = 0; s < nScenario; s++){
                for(int i = 0; i < order.size()+2; i++){
                    actual[s][i] = submodel.addVar(0, GRB.INFINITY, 0., GRB.CONTINUOUS, "wait"+s+","+i);
                }
            }
            GRBVar[][] wait = new GRBVar[nScenario][order.size()+1];
            for(int s = 0; s < nScenario; s++){
                for(int i = 0; i < order.size()+1; i++){
                    wait[s][i] = submodel.addVar(0, GRB.INFINITY, 0., GRB.CONTINUOUS, "wait"+s+","+i);
                }
            }
            GRBVar[] overtime = new GRBVar[nScenario];
            for(int s = 0; s < nScenario; s++){
                overtime[s] = submodel.addVar(0,GRB.INFINITY, 0, GRB.CONTINUOUS, "over"+s);
            }

            GRBLinExpr expr = new GRBLinExpr();
            for(int s = 0; s < nScenario; s++){
                expr.addTerm(0.1/nScenario, overtime[s]);
                for(int j = 1; j < order.size()+1; j++){
                    expr.addTerm(0.1/nScenario, wait[s][j]);
                }
            }
            submodel.setObjective(expr, GRB.MINIMIZE);

            for(int s = 0; s < nScenario; s++){
                for(int i = 0; i < order.size()+1; i++){
                    expr.clear();
                    expr.addTerm(1, actual[s][i]);
                    expr.addTerm(-1, gamma[i]);
                    expr.addTerm(-1, wait[s][i]);
                    submodel.addConstr(expr,GRB.EQUAL, 0, "wait_bind"+s+","+i);
                }

                expr.clear();
                expr.addTerm(1, actual[s][order.size()+1]);
                expr.addTerm(-1, overtime[s]);
                submodel.addConstr(expr, GRB.EQUAL, V1.get(vehicle).tw_b, "overtime_bind"+s);

                for(int i = 0; i < order.size()+1; i++){
                    expr.clear();
                    expr.addTerm(1, actual[s][i+1]);
                    expr.addTerm(-1, actual[s][i]);

                    double duration;
                    double trip;
                    if(i == 0){
                        duration = 0;
                        Node u = V0.get(vehicle);
                        Node v = V2.get(order.get(i));
                        trip = input.stochasticTravelMatrix[s][u.location][v.location];
                    }else {
                        Type2 temp = input.type2List.get(order.get(i-1));
                        duration = input.stochasticTravelMatrix[s][temp.origin][temp.destin];
                        Node u = V2.get(order.get(i-1));
                        Node v = i >= order.size()? V1.get(vehicle):V2.get(order.get(i));
                        trip = input.stochasticTravelMatrix[s][u.location][v.location];
                    }
                    submodel.addConstr(expr, GRB.GREATER_EQUAL, trip + duration, "connect"+s+i );
                }
            }

            submodel.optimize();
            double obj = submodel.get(GRB.DoubleAttr.ObjVal);
            ArrayList<Double> schedule_time = new ArrayList<>(order.size());
            double[] gammaSol = submodel.get(GRB.DoubleAttr.X, gamma);
            for(int i = 1; i< gammaSol.length-1; i++){
                schedule_time.add(gammaSol[i]);
            }
            submodel.dispose();
            return schedule_time;

        }catch (GRBException e) {
            System.out.println("Error code: " + e.getErrorCode() + ". " +
                    e.getMessage());
        }

        return new ArrayList<Double>();
    }

    public void solve_main(){
        System.out.println("start solving");


        try {
            String dirName = logDir + input.dirName;
            f = new File(dirName);
            if(!f.exists()){
                f.mkdir();
            }
            String logName = logDir + input.dirName + input.fileName + "_benders_new_"+nScenario+"_"+index;
            f = new File(logName);
            f.delete();
            GRBEnv env = new GRBEnv(logName);
            GRBModel  model = new GRBModel(env);
            model.set(GRB.DoubleParam.TimeLimit, 1800.0);
            //model.set(GRB.DoubleParam.MIPGap, 0.05);
            //model.getEnv().set(GRB.IntParam.DualReductions, 0);
            //model.set(GRB.IntParam.LazyConstraints, 1);


            // Create variables
            GRBVar[][] alpha = new GRBVar[E.size()][vehicle.size()];
            for(int i = 0; i < E0.size(); i++){
                for(int l = 0; l < vehicle.size(); l++){
                    if(E0.get(i).node1.idx_car == vehicle.get(l))
                        alpha[i][l] = model.addVar(0, 1, 0, GRB.BINARY, "alpha_"+i+"_"+l);
                    else{
                        alpha[i][l] = model.addVar(0, 0, 0, GRB.BINARY, "alpha_"+i+"_"+l);
                    }
                }

            }
            for(int i = E0.size(); i < E.size(); i++){
                for(int l = 0; l < vehicle.size(); l++){
                    alpha[i][l] = model.addVar(0, 1, 0, GRB.BINARY, "alpha_"+i+"_"+l);
                }

            }
            GRBVar[] beta = new GRBVar[nRide];
            for(int j = 0; j < nRide; j++){
                beta[j] = model.addVar(0, 1, 0, GRB.BINARY, "beta"+j);
            }

            GRBVar theta  = model.addVar(0,GRB.INFINITY,0, GRB.CONTINUOUS,"theta");


            GRBVar[] gamma = new GRBVar[V.size()];
            for(int i = 0; i < V.size(); i++){
                Node node = V.get(i);
                gamma[i] = model.addVar(node.tw_a, node.tw_b, 0, GRB.CONTINUOUS, "gamma"+i);
            }
            // Integrate new variables

            model.update();

            // Set objective: max visited node
            GRBLinExpr expr = new GRBLinExpr();
            for(int i = 0; i < nRide; i++){
                Type2 rr = input.type2List.get(i);
                expr.addTerm(ratePerHourRide / 60.0 * input.travelMatrix[rr.origin][rr.destin], beta[i]);
                //expr.addTerm(1, beta[i]);

            }
            expr.addTerm(-1,theta);
            model.setObjective(expr, GRB.MAXIMIZE);


            for(int j = 0; j < nRide; j++){
                expr.clear();
                for(int l = 0; l < vehicle.size(); l++){
                    for(Edge2 e : E0){
                        if(e.node2.idx_ride == j){
                            expr.addTerm(1, alpha[e.index][l]);
                        }
                    }
                    for(Edge2 e : E1){
                        if(e.node2.idx_ride == j){
                            expr.addTerm(1, alpha[e.index][l]);
                        }
                    }
                }
                model.addConstr(expr, GRB.EQUAL, beta[j], "beta"+j);
            }

            // depart constraints
            for(int l = 0; l < vehicle.size(); l++){
                expr.clear();
                for(Edge2 e : E0){
                    if(e.node1.idx_car == vehicle.get(l)){
                        expr.addTerm(1, alpha[e.index][l]);
                    }
                }
                model.addConstr(expr, GRB.LESS_EQUAL, 1, "initial"+l);
            }


            // end constraints
            for(int l = 0; l < vehicle.size(); l++){
                expr.clear();
                for(Edge2 e : E2){
                    if(e.node2.idx_car == vehicle.get(l)){
                        expr.addTerm(1, alpha[e.index][l]);
                    }
                }
                for(Edge2 e : E0){
                    if(e.node1.idx_car == vehicle.get(l)){
                        expr.addTerm(-1, alpha[e.index][l]);
                    }
                }
                model.addConstr(expr, GRB.EQUAL, 0, "end"+l);
            }

            //flow balance
            for(int l = 0; l < vehicle.size(); l++){
                for(Node v : V2){
                    expr.clear();
                    for(Edge2 e : E){
                        if(e.node2.idx_ride == v.idx_ride){
                            expr.addTerm(1, alpha[e.index][l]);
                        }else if(e.node1.idx_ride == v.idx_ride){
                            expr.addTerm(-1, alpha[e.index][l]);
                        }
                    }
                    model.addConstr(expr, GRB.EQUAL, 0, "flowbalance"+v.index+"-"+l);
                }
            }



            for(Edge2 e: E0){
                Node u = e.node1;
                Node v = e.node2;
                expr.clear();
                expr.addTerm(1,gamma[u.index]);
                expr.addTerm(-1, gamma[v.index]);
                for(int l = 0; l < vehicle.size(); l++){
                    if(vehicle.get(l) == u.idx_car){
                        expr.addTerm(12*60, alpha[e.index][l]);
                    }
                }
                model.addConstr(expr, GRB.LESS_EQUAL, 12*60 - input.travelMatrix[u.location][v.location], "time arrival"+e.index);
            }

            for(Edge2 e: E1){
                Node u = e.node1;
                Node v = e.node2;
                Type2 temp = input.type2List.get(u.idx_ride);
                expr.clear();
                expr.addTerm(1,gamma[u.index]);
                expr.addTerm(-1, gamma[v.index]);
                for(int l = 0; l < vehicle.size(); l++){
                    expr.addTerm(12*60, alpha[e.index][l]);
                }
                model.addConstr(expr, GRB.LESS_EQUAL, 12*60 - input.travelMatrix[temp.origin][temp.destin] -
                        input.travelMatrix[temp.destin][v.location], "time arrival"+e.index);
            }

            for(Edge2 e: E2){
                Node u = e.node1;
                Node v = e.node2;
                Type2 temp = input.type2List.get(u.idx_ride);
                expr.clear();
                expr.addTerm(1,gamma[u.index]);
                expr.addTerm(-1, gamma[v.index]);
                for(int l = 0; l < vehicle.size(); l++){
                    if (vehicle.get(l) == v.idx_car) {
                        expr.addTerm(12 * 60, alpha[e.index][l]);
                    }
                }
                model.addConstr(expr, GRB.LESS_EQUAL, 12*60 - input.travelMatrix[temp.origin][temp.destin] -
                        input.travelMatrix[temp.destin][v.location], "time arrival"+e.index);
            }

            Timer timer = new Timer();
            //model.setCallback(new Phase2Callback(alpha, beta, theta, gamma, timer));

            boolean flag = true;
            long overallStartTime = System.currentTimeMillis();
            while(flag){
                {
                    flag = false;
                    model.optimize();
                    System.out.println("CURRENT OBJ:"+model.get(GRB.DoubleAttr.ObjVal));
                    double boundObj = model.get(GRB.DoubleAttr.ObjVal);
                    if(boundObj < upperBound){
                        upperBound = boundObj;
                    }
                    System.out.println("CURRENT UB, LB" + upperBound + "," + lowerBound);
                    double[][] solution_alpha = model.get(GRB.DoubleAttr.X, alpha);
                    double[] solution_gamma = model.get(GRB.DoubleAttr.X, gamma);
                    double[] solution_beta = model.get(GRB.DoubleAttr.X, beta);
                    long start = System.currentTimeMillis();
                    System.out.println("CURRENT GAMMA:"+Arrays.toString(solution_gamma));
                    System.out.println("CURRENT BETA:"+Arrays.toString(solution_beta));


                    double lb_theta = 0;


                    double cur_theta = theta.get(GRB.DoubleAttr.X);
                    double[][] pi = new double[300][E0.size()];
                    double[][] mu = new double[300][E1.size()];
                    double[][] nu = new double[300][E2.size()];



                    for(int s = 0 ; s < 300; s++) {
                        GRBEnv subenv = new GRBEnv();
                        GRBModel  sub = new GRBModel(subenv);
                        GRBVar[] wait = new GRBVar[nRide];
                        for (int j = 0; j < nRide; j++) {
                            wait[j] = sub.addVar(0, GRB.INFINITY, 0, GRB.CONTINUOUS, "wait" + j + "_" + s);
                        }


                        GRBVar[] overtime = new GRBVar[nCar];
                        for (int l = 0; l < nCar; l++) {
                            if (vehicle.contains(l)) {
                                overtime[l] = sub.addVar(0, GRB.INFINITY, 0, GRB.CONTINUOUS, "overtime" + l + "_" + s);
                            } else {
                                overtime[l] = sub.addVar(0, 0, 0, GRB.CONTINUOUS, "overtime" + l + "_" + s);
                            }
                        }

                        sub.update();
                        expr = new GRBLinExpr();
                        for(int l : vehicle){
                            expr.addTerm(0.1, overtime[l]);
                        }
                        for(int j = 0; j < nRide; j++){
                            expr.addTerm(0.1, wait[j]);
                        }

                        sub.setObjective(expr, GRB.MINIMIZE);
                        // compute wait/idle time
                        GRBConstr[] cpi = new GRBConstr[E0.size()];
                        int idx = 0;
                        for (Edge2 e : E0) {
                            Node u = e.node1;
                            Node v = e.node2;
                            expr.clear();
                            double rhs = 0;
                            rhs += solution_gamma[u.index]- solution_gamma[v.index];
                            expr.addTerm(-1, wait[v.idx_ride]);
                            for (int l = 0; l < vehicle.size(); l++) {
                                if (vehicle.get(l) == u.idx_car) {
                                    rhs+= 12 * 60 * solution_alpha[e.index][l];
                                }
                            }
                            //					System.out.println(Arrays.toString(ins.distance_s[9][0]));
                            cpi[idx] = sub.addConstr(expr, GRB.LESS_EQUAL, 12 * 60 - input.stochasticTravelMatrix[s][u.location][v.location] - rhs, "time arrival" + e.index);
                            idx++;
                        }

                        GRBConstr[] cmu = new GRBConstr[E1.size()];
                        idx = 0;
                        for (Edge2 e : E1) {
                            Node u = e.node1;
                            Node v = e.node2;
                            Type2 temp = input.type2List.get(u.idx_ride);
                            expr.clear();
                            double rhs = 0;
                            rhs += solution_gamma[u.index] - solution_gamma[v.index];
                            expr.addTerm(1, wait[u.idx_ride]);
                            //expr.addTerm(-1, idle[o.idx_ride][s]);
                            expr.addTerm(-1, wait[v.idx_ride]);
                            //expr.addTerm(1, idle[v.idx_ride][s]);
                            for (int l = 0; l < vehicle.size(); l++) {
                                rhs += 12 * 60 *solution_alpha[e.index][l];
                            }
                            cmu[idx] = sub.addConstr(expr, GRB.LESS_EQUAL, 12 * 60 - input.stochasticTravelMatrix[s][temp.origin][temp.destin] - input.stochasticTravelMatrix[s][temp.destin][v.location] - rhs, "time arrival" + e.index);
                            idx++;
                        }


                        GRBConstr[] cnu = new GRBConstr[E2.size()];
                        idx = 0;
                        for (Edge2 e : E2) {
                            Node u = e.node1;
                            Node v = e.node2;
                            Type2 temp = input.type2List.get(u.idx_ride);
                            expr.clear();
                            double rhs = 0;
                            rhs += solution_gamma[u.index] - solution_gamma[v.index];

                            expr.addTerm(1, wait[u.idx_ride]);
                            expr.addTerm(-1, overtime[v.idx_car]);
                            for (int l = 0; l < vehicle.size(); l++) {
                                if (vehicle.get(l) == v.idx_car) {
                                    rhs += (12 * 60 * solution_alpha[e.index][l]);
                                }
                            }
                            cnu[idx] = sub.addConstr(expr, GRB.LESS_EQUAL, 12 * 60 - input.stochasticTravelMatrix[s][temp.origin][temp.destin] - input.stochasticTravelMatrix[s][temp.destin][v.location] - rhs, "time arrival" + e.index);
                            idx++;
                        }
                        sub.optimize();
                        double obj = sub.get(GRB.DoubleAttr.ObjVal);

                        double[] solution_pi = sub.get(GRB.DoubleAttr.Pi, cpi);
                        double[] solution_mu = sub.get(GRB.DoubleAttr.Pi, cmu);
                        double[] solution_nu = sub.get(GRB.DoubleAttr.Pi, cnu);

                        System.arraycopy(solution_pi,0,pi[s],0,E0.size());
                        System.arraycopy(solution_mu,0,mu[s],0,E1.size());
                        System.arraycopy(solution_nu,0,nu[s],0,E2.size());

                        lb_theta+= obj;

                        System.out.println("subproblem #"+s);
                        //System.out.println("PI"+Arrays.toString(mu[271]));
                        sub = null;
                    }

                    System.out.println("Finished solving subproblems!");


                    //System.out.println(sum_cur);
                    //System.out.println(sum_lb+"lb");
                    GRBLinExpr ex = new GRBLinExpr();
                    double rhs = 0;
                    System.out.println("sub_theta: "+(lb_theta/300.0)+" vs. "+ cur_theta);

                    if(boundObj - lb_theta/300.0 > lowerBound){
                        System.out.println("Find new lower bound!");
                        lowerBound = boundObj - lb_theta/300.0;
                        bestGammaSol = solution_gamma;
                        bestBetaSol = solution_beta;
                        bestAlphaSol = solution_alpha;
                        System.out.println(Arrays.toString(bestBetaSol));
                    }



                    if(lb_theta/300.0 > cur_theta){
                        flag = true;
                        timer.counter++;
                        timer.counter_sub++;
                        ex.clear();
                        int idx = 0;
                        for(int s = 0; s < 300; s++){
                            idx = 0;
                            for (Edge2 e : E0) {
                                Node u = e.node1;
                                Node v = e.node2;
                                ex.addTerm(1.0/300 *-pi[s][idx], gamma[u.index]);
                                ex.addTerm(-1.0/300 *-pi[s][idx], gamma[v.index]);
                                for (int l = 0; l < vehicle.size(); l++) {
                                    if (vehicle.get(l) == u.idx_car) {
                                        ex.addTerm(1.0/300 *12 * 60 * -pi[s][idx], alpha[e.index][l]);
                                    }
                                }
                                rhs += (12 * 60 - input.stochasticTravelMatrix[s][u.location][v.location])*-pi[s][idx]/300;

                                idx++;
                            }

                            idx = 0;
                            for (Edge2 e : E1) {
                                Node u = e.node1;
                                Node v = e.node2;
                                Type2 temp = input.type2List.get(u.idx_ride);
                                ex.addTerm(1.0/300 *-mu[s][idx], gamma[u.index]);
                                ex.addTerm(-1.0/300 *-mu[s][idx], gamma[v.index]);
                                for (int l = 0; l < vehicle.size(); l++) {
                                    ex.addTerm(1.0/300 *12 * 60 * -mu[s][idx], alpha[e.index][l]);
                                }
                                rhs += (12 * 60 - input.stochasticTravelMatrix[s][temp.origin][temp.destin] - input.stochasticTravelMatrix[s][temp.destin][v.location])*-mu[s][idx]/300;
                                idx++;
                            }


                            idx = 0;
                            for (Edge2 e : E2) {
                                Node u = e.node1;
                                Node v = e.node2;
                                Type2 temp = input.type2List.get(u.idx_ride);
                                ex.addTerm(1.0/300 *-nu[s][idx], gamma[u.index]);
                                ex.addTerm(-1.0/300 *-nu[s][idx], gamma[v.index]);
                                for (int l = 0; l < vehicle.size(); l++) {
                                    if (vehicle.get(l) == v.idx_car) {
                                        ex.addTerm(1.0 / 300 * 12 * 60 * -nu[s][idx], alpha[e.index][l]);
                                    }
                                }

                                rhs += (12 * 60 - input.stochasticTravelMatrix[s][temp.origin][temp.destin] - input.stochasticTravelMatrix[s][temp.destin][v.location])*-nu[s][idx]/300;
                                idx++;
                            }

                        }
                        ex.addTerm(-1, theta);

                        System.out.println("adding cut!");
                        System.out.println("CURRENT RHS: "+rhs);
                        model.addConstr(ex, GRB.LESS_EQUAL, rhs, "cut");
                        model.update();


                    }

                    timer.time += (System.currentTimeMillis()- start);
                    if(System.currentTimeMillis() - overallStartTime > 1800000){
                        timer.overallTime = System.currentTimeMillis() - overallStartTime;
                        flag = false;
                    }
                }
            }

            model.optimize();
            System.out.println("Solved!");
            System.out.println(alpha.length);
            double runtime = timer.overallTime;//model.get(GRB.DoubleAttr.Runtime);
            result.runtime_phaseII_stochastic = (double) Math.round(runtime * 1000d) / 1000000d;

            double[][] solution_alpha = bestAlphaSol;//model.get(GRB.DoubleAttr.X, alpha);
            //System.out.println("Solution alpha:"+Arrays.toString(solution_alpha[799]));
            result.rawSolution_alpha = solution_alpha;


            double[] solution_beta = bestBetaSol;//model.get(GRB.DoubleAttr.X, beta);
            System.out.println("Solution beta:"+Arrays.toString(solution_beta));
            result.rawSolution_beta = solution_beta;
            int count = 0;
            for(int i = 0; i < nRide; i++){
                if(solution_beta[i]  > 0.5){
                    count += 1;
                }
            }
            result.fulfillmentRate_s_Ride = ((double) count )/ nRide;

            ArrayList<ArrayList<Integer>> arrangement = new ArrayList<>();
            for(int i = 0; i < nCar; i++){
                arrangement.add(new ArrayList<>());
            }


            for(Edge2 e : E0){
                for(int l = 0; l < vehicle.size(); l++){
                    if(solution_alpha[e.index][l] > 0.5){
                        //System.out.println("Ride "+e.node2.idx_ride+" is served by "+e.node1.idx_car+"("+l+")" );
                        arrangement.get(vehicle.get(l)).add(e.node2.idx_ride);
                    }
                }
            }
            for(int l = 0; l < vehicle.size(); l++){
                flag = true;
                while (flag){
                    flag = false;
                    if(arrangement.get(vehicle.get(l)).size() < 1)
                        continue;
                    int temp = arrangement.get(vehicle.get(l)).get(arrangement.get(vehicle.get(l)).size()-1);
                    for(Edge2 e : E1){
                        if(solution_alpha[e.index][l] > 0.5 && e.node1.idx_ride == temp){
                            //System.out.println("Ride "+e.node2.idx_ride+" is served by "+e.node1.idx_car+"("+l+")" );
                            arrangement.get(vehicle.get(l)).add(e.node2.idx_ride);
                            flag = true;
                            System.out.println(vehicle.get(l)+","+temp);

                        }
                    }
                }

            }

            for(int i = 0; i < nCar; i++){
                System.out.println("arrangement for car "+i+": "+arrangement.get(i).toString());
//                if(arrangement.get(i).size() > 0)
//                    System.out.println(subproblem(arrangement.get(i), i));
            }

            //System.out.println(Arrays.toString(newArray));
            result.assignment = arrangement;
            result.rawSolution_s_gamma = new double[nRide];
            for(int i = 0; i < vehicle.size(); i++){
                if(arrangement.get(vehicle.get(i)).size() <= 0)
                    continue;
                ArrayList<Double> temp = subproblem2(arrangement.get(vehicle.get(i)), i);
                for(int j= 0 ; j < arrangement.get(vehicle.get(i)).size(); j++){
                    System.out.println(temp.get(j));
                    result.rawSolution_s_gamma[arrangement.get(vehicle.get(i)).get(j)] = temp.get(j);
                }
            }
            //System.out.println(timer.counter+","+timer.counter_sub+","+timer.time);
            result.callback_count = timer.counter;
            result.subproblem_time = timer.time;
            result.subproblem_count = timer.counter_sub;
//            System.out.println("Obj: " + model.get(GRB.DoubleAttr.ObjVal));
            double upperbound = upperBound;
            double lowerbound = lowerBound;
            result.gap_s_phaseII = 1- lowerbound/upperbound;
            // Dispose of model and environment
            result.outputPhaseIIBenders();
            model.dispose();
            env.dispose();

        } catch (GRBException e) {
            System.out.println("Error code: " + e.getErrorCode() + ". " +
                    e.getMessage());
        }
    }

}