package solver;

import gurobi.*;

import java.io.File;
import java.util.*;

import static helper.Global.*;
import structure.Structures.*;

public class Solver {
    public Input input;
    public Log result;
    public File f;
//    public double ratePerHourCar = 8;
//    public double ratePerHourRide = 30;
//    public double costIdle = 0;
//    public double costService = 0;
//    public double rateDriveOther = 15;


    public Solver(Input i){
        this.input = i;
        this.result = new Log(input);
    }


    public void solvePhaseI(){
        // prepare data
        int nZone = input.nSite;
        Map<Integer, Integer> location = new HashMap<Integer, Integer>();
        int nPeriod = 12;
        ArrayList<ArrayList<ArrayList<Integer>>> adj = new ArrayList<ArrayList<ArrayList<Integer>>>();
        for(int i = 0; i < nZone; i++){
            adj.add(new ArrayList<ArrayList<Integer>>());
            for(int t = 0; t < nPeriod + 1; t++){
                adj.get(i).add(new ArrayList<Integer>());
            }
        }
        ArrayList<ArrayList<ArrayList<Integer>>> inj = new ArrayList<ArrayList<ArrayList<Integer>>>();
        for(int i = 0; i < nZone; i++){
            inj.add(new ArrayList<ArrayList<Integer>>());
            for(int t = 0; t < nPeriod + 1; t++){
                inj.get(i).add(new ArrayList<Integer>());
            }
        }

        List<Edge> edgeList = new ArrayList<Edge>();
        List<Edge> tripEdge = new ArrayList<Edge>();
        int counter = 0;

        // create traveling edges
        for(Type1 cr : input.type1List){
            Edge e = new Edge();
            int l1, l2;
            if(location.containsKey(cr.origin)){
                l1 = location.get(cr.origin);
            }else{
                location.put(cr.origin, location.keySet().size());
                l1 = location.get(cr.origin);
            }
            if(location.containsKey(cr.destin)){
                l2 = location.get(cr.destin);
            }else{
                location.put(cr.destin, location.keySet().size());
                l2 = location.get(cr.destin);
            }
            e.loc1 = l1;
            e.loc2 = l2;
            e.t1 = cr.start;
            e.t2 = cr.end;
            e.idx = counter;
            e.cost = (e.t2 - e.t1) * (ratePerHourCar - costService);
            e.type = 1;
            edgeList.add(e);
            tripEdge.add(e);
            adj.get(l1).get(e.t1).add(e.idx);
            inj.get(l2).get(e.t2).add(e.idx);
            counter++;
        }
        // create idle arcs
        for(int i = 0; i < nZone; i++){
            for(int t = 0; t < nPeriod; t++){
                Edge e = new Edge();
                e.loc1 = i;
                e.loc2 = i;
                e.t1 = t;
                e.t2 = t+1;
                e.cost = -costIdle;
                e.idx = counter;
                e.type = 2;
                edgeList.add(e);
                adj.get(i).get(t).add(e.idx);
                inj.get(i).get(t+1).add(e.idx);
                counter++;
            }
        }
        // compute w
        int nCar = input.nType1;
        int nRide = input.nType2;
        int nCars = input.nVehicle;
        int[][] w = new int[nRide][nCar];
        for(int j = 0; j < nRide; j++)
            for(int l = 0; l < nCar; l++){
                Type1 car = input.type1List.get(l);
                Type2 ride = input.type2List.get(j);
                if((ride.tw_a >= car.end * 60)||(ride.tw_b <= car.startD * 60)){
                    w[j][l] = 0;
                }else if(ride.tw_a <= car.startD * 60){
                    if(car.startD * 60 + input.travelMatrix[ride.origin][ride.destin] <= car.end * 60){
                        w[j][l] = 1;
                    }else{
                        w[j][l] = 0;
                    }
                }else{
                    if(ride.tw_a + input.travelMatrix[ride.origin][ride.destin] <= car.end * 60){
                        w[j][l] = 1;
                    }else{
                        w[j][l] = 0;
                    }
                }
            }

        try{
            //Model
            String dirName = logDir + input.dirName;
            f = new File(dirName);
            if(!f.exists()){
                f.mkdir();
            }
            String logName = logDir + input.dirName+ input.fileName +"_gurobi_phase1";
            f = new File(logName);
            f.delete();

            GRBEnv env = new GRBEnv(logName);
            GRBModel model = new GRBModel(env);
            model.set(GRB.DoubleParam.TimeLimit, 1800.0);
            GRBVar[] x = new GRBVar[nZone];
            for(int i = 0; i < nZone; i++){
                x[i] = model.addVar(0, nCars, 0, GRB.INTEGER, "x"+i);
            }
            //			System.out.println(nCars);
            GRBVar[] y = new GRBVar[edgeList.size()];
            for(int i = 0; i < edgeList.size(); i++){
                if(edgeList.get(i).type == 1){
                    y[i] = model.addVar(0, 1, 0, GRB.BINARY, "y"+i);
                }else if(edgeList.get(i).type == 2){
                    y[i] = model.addVar(0, nCars, 0, GRB.INTEGER, "y"+i);
                }
            }

            GRBVar[][] z = new GRBVar[input.type2List.size()][input.type1List.size()];
            for(int j = 0; j < input.type2List.size(); j++){
                for(int l = 0; l < input.type1List.size(); l++){
                    z[j][l] = model.addVar(0, w[j][l], 0, GRB.BINARY, "z-"+j+"-"+l);
                }
            }

            model.update();

            GRBLinExpr expr = new GRBLinExpr();
            for(Edge e : tripEdge){
                expr.addTerm(1, y[e.idx]);
            }
            for(int j = 0; j < input.type2List.size(); j++){
                for(int l = 0; l < input.type1List.size(); l++){
                    expr.addTerm(1, z[j][l]);
                }
            }
            model.setObjective(expr, GRB.MAXIMIZE);

            expr.clear();
            for(int i = 0 ; i < nZone; i++){
                expr.addTerm(1, x[i]);
            }
            model.addConstr(expr, GRB.EQUAL, nCars, "initial location");

            for(int i = 0; i < nZone; i++){
                expr.clear();
                for(int edge : adj.get(i).get(0)){
                    expr.addTerm(1, y[edge]);
                }
                expr.addTerm(-1, x[i]);
                model.addConstr(expr, GRB.EQUAL, 0, "initial flow "+i);
            }

            for(int t = 1; t < nPeriod; t++){
                for(int i = 0; i < nZone; i++){
                    expr.clear();
                    for(int edge : adj.get(i).get(t)){
                        expr.addTerm(1, y[edge]);
                    }
                    for(int edge: inj.get(i).get(t)){
                        expr.addTerm(-1, y[edge]);

                    }
                    model.addConstr(expr, GRB.EQUAL, 0, "flow balance"+"-"+t);
                }
            }

            for(int i = 0; i < nZone; i++){
                expr.clear();
                for(int edge : inj.get(i).get(nPeriod)){
                    expr.addTerm(1, y[edge]);
                }
                expr.addTerm(-1, x[i]);
                model.addConstr(expr, GRB.EQUAL, 0, "initial flow "+i);
            }

            for(int l = 0; l < nCar; l++){
                Type1 car = input.type1List.get(l);
                expr.clear();
                for(int j = 0; j < nRide; j++){
                    Type2 ride = input.type2List.get(j);
                    expr.addTerm(input.travelMatrix[ride.origin][ride.destin], z[j][l]);
                }
                expr.addTerm(-60 * (car.end - car.startD), y[l]);
                model.addConstr(expr, GRB.LESS_EQUAL, 0, "matching"+l);
            }

            for(int j = 0; j < nRide; j++){
                expr.clear();
                for(int l = 0; l < nCar; l++){
                    expr.addTerm(1, z[j][l]);
                }
                model.addConstr(expr, GRB.LESS_EQUAL, 1, "matched"+j);
            }

            model.update();
            model.optimize();

            if(model.get(GRB.IntAttr.Status) != GRB.Status.OPTIMAL){
                System.err.println("Error: model is infeasible");
                return;
            }
            double lowerbound = model.get(GRB.DoubleAttr.ObjBound);
            double upperbound = model.get(GRB.DoubleAttr.ObjVal);
            result.gap_phaseI = 1- lowerbound/upperbound;


            double[] solution_x = model.get(GRB.DoubleAttr.X, x);
//            System.out.println("Solution x:"+Arrays.toString(solution_x));
            result.rawSolution_x = solution_x;
            double[] solution_y = model.get(GRB.DoubleAttr.X, y);
            result.rawSolution_y = solution_y;

            double[][] idle = new double[nZone][nPeriod];
            for(int i = 0; i < nZone; i++){
                for(int t = 0; t < nPeriod; t++){
                    for(int idx : adj.get(i).get(t)){
                        if( edgeList.get(idx).type ==  2){
                            //System.out.println("["+i+","+t+"]: "+solution_y[idx]);
                            idle[i][t] = solution_y[idx];
                        }
                    }
                }
            }
            result.idleness = idle;
            int count = 0;
            for(int l = 0 ; l < nCar; l++){
                if(solution_y[l] >= 1-1E-6){
                    input.type1List.get(l).approved = true;
                    count += 1;
                }
            }
            result.fulfillmentRate_Car = ((double) count) / nCar;
//            System.out.println("Solution y:"+Arrays.toString(solution_y));
            double runtime = model.get(GRB.DoubleAttr.Runtime);
            result.runtime_phaseI = (double) Math.round(runtime * 1000d) / 1000d;
            result.outputPhaseI();
            model.dispose();
            env.dispose();

        } catch (GRBException e) {
            System.out.println("Error code: " + e.getErrorCode() + ". " +
                    e.getMessage());
        }
    }

    public void solvePhaseI_relocation(){
        // prepare data
        int nZone = input.nSite;
        Map<Integer, Integer> location = new HashMap<Integer, Integer>();
        int nPeriod = 12;
        ArrayList<ArrayList<ArrayList<Integer>>> adj = new ArrayList<ArrayList<ArrayList<Integer>>>();
        for(int i = 0; i < nZone; i++){
            adj.add(new ArrayList<ArrayList<Integer>>());
            for(int t = 0; t < nPeriod + 1; t++){
                adj.get(i).add(new ArrayList<Integer>());
            }
        }
        ArrayList<ArrayList<ArrayList<Integer>>> inj = new ArrayList<ArrayList<ArrayList<Integer>>>();
        for(int i = 0; i < nZone; i++){
            inj.add(new ArrayList<ArrayList<Integer>>());
            for(int t = 0; t < nPeriod + 1; t++){
                inj.get(i).add(new ArrayList<Integer>());
            }
        }

        List<Edge> edgeList = new ArrayList<Edge>();
        List<Edge> tripEdge = new ArrayList<Edge>();
        int counter = 0;

        // create traveling edges
        for(Type1 cr : input.type1List){
            Edge e = new Edge();
            int l1, l2;
            if(location.containsKey(cr.origin)){
                l1 = location.get(cr.origin);
            }else{
                location.put(cr.origin, location.keySet().size());
                l1 = location.get(cr.origin);
            }
            if(location.containsKey(cr.destin)){
                l2 = location.get(cr.destin);
            }else{
                location.put(cr.destin, location.keySet().size());
                l2 = location.get(cr.destin);
            }
            e.loc1 = l1;
            e.loc2 = l2;
            e.t1 = cr.start;
            e.t2 = cr.end;
            e.idx = counter;
            e.cost = (e.t2 - e.t1) * (ratePerHourCar - costService);
            e.type = 1;
            edgeList.add(e);
            tripEdge.add(e);
            adj.get(l1).get(e.t1).add(e.idx);
            inj.get(l2).get(e.t2).add(e.idx);
            counter++;
        }
        // create idle arcs
        for(int i = 0; i < nZone; i++){
            for(int t = 0; t < nPeriod; t++){
                Edge e = new Edge();
                e.loc1 = i;
                e.loc2 = i;
                e.t1 = t;
                e.t2 = t+1;
                e.cost = -costIdle;
                e.idx = counter;
                e.type = 2;
                edgeList.add(e);
                adj.get(i).get(t).add(e.idx);
                inj.get(i).get(t+1).add(e.idx);
                counter++;
            }
        }

        // build array for locations
        int[] locs = new int[location.size()];
        for(int num : location.keySet()){
            locs[location.get(num)] = num;
        }
        System.out.println(Arrays.toString(locs));
        System.out.println("before: "+counter);

        // create relocation arcs
        for(int i = 0; i < locs.length; i++){
            for(int t = 0; t < nPeriod; t++){
                for(int j = 0; j < locs.length; j++){
                    if(i == j)
                        continue;
                    Edge e = new Edge();
                    e.loc1 = i;
                    e.loc2 = j;
                    e.t1 = t;
                    e.t2 = t+1;
                    e.cost =  -input.travelMatrix[locs[i]][locs[j]] / 60.0 *  costRelocation;
                    e.type = 3;
                    e.idx = counter;
                    edgeList.add(e);
                    adj.get(i).get(t).add(e.idx);
                    inj.get(j).get(t+1).add(e.idx);
                    counter++;
                }
            }
        }
        System.out.println("after: "+counter);
        // compute w
        int nCar = input.nType1;
        int nRide = input.nType2;
        int nCars = input.nVehicle;
        int[][] w = new int[nRide][nCar];
        for(int j = 0; j < nRide; j++)
            for(int l = 0; l < nCar; l++){
                Type1 car = input.type1List.get(l);
                Type2 ride = input.type2List.get(j);
                if((ride.tw_a >= car.end * 60)||(ride.tw_b <= car.startD * 60)){
                    w[j][l] = 0;
                }else if(ride.tw_a <= car.startD * 60){
                    if(car.startD * 60 + input.travelMatrix[ride.origin][ride.destin] <= car.end * 60){
                        w[j][l] = 1;
                    }else{
                        w[j][l] = 0;
                    }
                }else{
                    if(ride.tw_a + input.travelMatrix[ride.origin][ride.destin] <= car.end * 60){
                        w[j][l] = 1;
                    }else{
                        w[j][l] = 0;
                    }
                }
            }

        try{
            //Model
            String dirName = logDir + input.dirName;
            f = new File(dirName);
            if(!f.exists()){
                f.mkdir();
            }
            String logName = logDir + input.dirName+ input.fileName +"_gurobi_phase1_rel";
            f = new File(logName);
            f.delete();

            GRBEnv env = new GRBEnv(logName);
            GRBModel model = new GRBModel(env);
            model.set(GRB.DoubleParam.TimeLimit, 1800.0);
            GRBVar[] x = new GRBVar[nZone];
            for(int i = 0; i < nZone; i++){
                x[i] = model.addVar(0, nCars, 0, GRB.INTEGER, "x"+i);
            }
            //			System.out.println(nCars);
            GRBVar[] y = new GRBVar[edgeList.size()];
            for(int i = 0; i < edgeList.size(); i++){
                if(edgeList.get(i).type == 1){
                    y[i] = model.addVar(0, 1, 0, GRB.BINARY, "y"+i);
                }else {
                    y[i] = model.addVar(0, nCars, 0, GRB.INTEGER, "y"+i);
                }
            }

            GRBVar[][] z = new GRBVar[input.type2List.size()][input.type1List.size()];
            for(int j = 0; j < input.type2List.size(); j++){
                for(int l = 0; l < input.type1List.size(); l++){
                    z[j][l] = model.addVar(0, w[j][l], 0, GRB.BINARY, "z-"+j+"-"+l);
                }
            }

            model.update();

            GRBLinExpr expr = new GRBLinExpr();
//            for(Edge e : tripEdge){
//                expr.addTerm(1, y[e.idx]);
//            }
            for(Edge e : edgeList){
                expr.addTerm(e.cost, y[e.idx]);
            }
            for(int j = 0; j < input.type2List.size(); j++){
                for(int l = 0; l < input.type1List.size(); l++){
                    expr.addTerm(1, z[j][l]);
                }
            }
            model.setObjective(expr, GRB.MAXIMIZE);

            expr.clear();
            for(int i = 0 ; i < nZone; i++){
                expr.addTerm(1, x[i]);
            }
            model.addConstr(expr, GRB.EQUAL, nCars, "initial location");

            for(int i = 0; i < nZone; i++){
                expr.clear();
                for(int edge : adj.get(i).get(0)){
                    expr.addTerm(1, y[edge]);
                }
                expr.addTerm(-1, x[i]);
                model.addConstr(expr, GRB.EQUAL, 0, "initial flow "+i);
            }

            for(int t = 1; t < nPeriod; t++){
                for(int i = 0; i < nZone; i++){
                    expr.clear();
                    for(int edge : adj.get(i).get(t)){
                        expr.addTerm(1, y[edge]);
                    }
                    for(int edge: inj.get(i).get(t)){
                        expr.addTerm(-1, y[edge]);

                    }
                    model.addConstr(expr, GRB.EQUAL, 0, "flow balance"+"-"+t);
                }
            }

//            for(int i = 0; i < nZone; i++){
//                expr.clear();
//                for(int edge : inj.get(i).get(nPeriod)){
//                    expr.addTerm(1, y[edge]);
//                }
//                expr.addTerm(-1, x[i]);
//                model.addConstr(expr, GRB.EQUAL, 0, "terminal flow "+i);
//            }

            for(int l = 0; l < nCar; l++){
                Type1 car = input.type1List.get(l);
                expr.clear();
                for(int j = 0; j < nRide; j++){
                    Type2 ride = input.type2List.get(j);
                    expr.addTerm(input.travelMatrix[ride.origin][ride.destin], z[j][l]);
                }
                expr.addTerm(-60 * (car.end - car.startD), y[l]);
                model.addConstr(expr, GRB.LESS_EQUAL, 0, "matching"+l);
            }

            for(int j = 0; j < nRide; j++){
                expr.clear();
                for(int l = 0; l < nCar; l++){
                    expr.addTerm(1, z[j][l]);
                }
                model.addConstr(expr, GRB.LESS_EQUAL, 1, "matched"+j);
            }

            model.update();
            model.optimize();

            if(model.get(GRB.IntAttr.Status) != GRB.Status.OPTIMAL){
                System.err.println("Error: model is infeasible");
                return;
            }
            double lowerbound = model.get(GRB.DoubleAttr.ObjBound);
            double upperbound = model.get(GRB.DoubleAttr.ObjVal);
            result.gap_phaseI = 1- lowerbound/upperbound;


            double[] solution_x = model.get(GRB.DoubleAttr.X, x);
//            System.out.println("Solution x:"+Arrays.toString(solution_x));
            result.rawSolution_x = solution_x;
            double[] solution_y = model.get(GRB.DoubleAttr.X, y);
            result.rawSolution_y = solution_y;

            double[][] idle = new double[nZone][nPeriod];
            for(int i = 0; i < nZone; i++){
                for(int t = 0; t < nPeriod; t++){
                    for(int idx : adj.get(i).get(t)){
                        if( edgeList.get(idx).type ==  2){
                            //System.out.println("["+i+","+t+"]: "+solution_y[idx]);
                            idle[i][t] = solution_y[idx];
                        }
                    }
                }
            }
            result.idleness = idle;

            double[][][] relocation = new double[nPeriod][nZone][nZone];
            for(int i = 0; i < nZone; i++){
                for(int t = 0; t < nPeriod; t++){
                    for(int idx : adj.get(i).get(t)){
                        if(edgeList.get(idx).type == 3){
                            relocation[t][i][edgeList.get(idx).loc2] = solution_y[idx];
                        }
                    }
                }
            }
            result.relocation = relocation;


            int count = 0;
            for(int l = 0 ; l < nCar; l++){
                if(solution_y[l] >= 1-1E-6){
                    input.type1List.get(l).approved = true;
                    count += 1;
                }
            }
            result.fulfillmentRate_Car = ((double) count) / nCar;
//            System.out.println("Solution y:"+Arrays.toString(solution_y));
            double runtime = model.get(GRB.DoubleAttr.Runtime);
            result.runtime_phaseI = (double) Math.round(runtime * 1000d) / 1000d;
            result.outputPhaseIRelocation();
            model.dispose();
            env.dispose();

        } catch (GRBException e) {
            System.out.println("Error code: " + e.getErrorCode() + ". " +
                    e.getMessage());
        }
    }

    public void integrated(){
        // prepare data
        int nZone = input.nSite;
        Map<Integer, Integer> location = new HashMap<Integer, Integer>();
        int nPeriod = 12;
        ArrayList<ArrayList<ArrayList<Integer>>> adj = new ArrayList<ArrayList<ArrayList<Integer>>>();
        for(int i = 0; i < nZone; i++){
            adj.add(new ArrayList<ArrayList<Integer>>());
            for(int t = 0; t < nPeriod + 1; t++){
                adj.get(i).add(new ArrayList<Integer>());
            }
        }
        ArrayList<ArrayList<ArrayList<Integer>>> inj = new ArrayList<ArrayList<ArrayList<Integer>>>();
        for(int i = 0; i < nZone; i++){
            inj.add(new ArrayList<ArrayList<Integer>>());
            for(int t = 0; t < nPeriod + 1; t++){
                inj.get(i).add(new ArrayList<Integer>());
            }
        }

        List<Edge> edgeList = new ArrayList<Edge>();
        List<Edge> tripEdge = new ArrayList<Edge>();
        int counter = 0;

        // create traveling edges
        for(Type1 cr : input.type1List){
            Edge e = new Edge();
            int l1, l2;
            if(location.containsKey(cr.origin)){
                l1 = location.get(cr.origin);
            }else{
                location.put(cr.origin, location.keySet().size());
                l1 = location.get(cr.origin);
            }
            if(location.containsKey(cr.destin)){
                l2 = location.get(cr.destin);
            }else{
                location.put(cr.destin, location.keySet().size());
                l2 = location.get(cr.destin);
            }
            e.loc1 = l1;
            e.loc2 = l2;
            e.t1 = cr.start;
            e.t2 = cr.end;
            e.idx = counter;
            e.cost = (e.t2 - e.t1) * (ratePerHourCar - costService);
            e.type = 1;
            edgeList.add(e);
            tripEdge.add(e);
            adj.get(l1).get(e.t1).add(e.idx);
            inj.get(l2).get(e.t2).add(e.idx);
            counter++;
        }
        // create idle arcs
        for(int i = 0; i < nZone; i++){
            for(int t = 0; t < nPeriod; t++){
                Edge e = new Edge();
                e.loc1 = i;
                e.loc2 = i;
                e.t1 = t;
                e.t2 = t+1;
                e.cost = -costIdle;
                e.idx = counter;
                e.type = 2;
                edgeList.add(e);
                adj.get(i).get(t).add(e.idx);
                inj.get(i).get(t+1).add(e.idx);
                counter++;
            }
        }

        // build array for locations
        int[] locs = new int[location.size()];
        for(int num : location.keySet()){
            locs[location.get(num)] = num;
        }
        System.out.println(Arrays.toString(locs));
        System.out.println("before: "+counter);

        // create relocation arcs
        for(int i = 0; i < locs.length; i++){
            for(int t = 0; t < nPeriod; t++){
                for(int j = 0; j < locs.length; j++){
                    if(i == j)
                        continue;
                    Edge e = new Edge();
                    e.loc1 = i;
                    e.loc2 = j;
                    e.t1 = t;
                    e.t2 = t+1;
                    e.cost =  -input.travelMatrix[locs[i]][locs[j]] / 60.0 *  costRelocation;
                    e.type = 3;
                    e.idx = counter;
                    edgeList.add(e);
                    adj.get(i).get(t).add(e.idx);
                    inj.get(j).get(t+1).add(e.idx);
                    counter++;
                }
            }
        }
        System.out.println("after: "+counter);
        // compute w
        int nCar = input.nType1;
        int nRide = input.nType2;
        int nCars = input.nVehicle;
//        int[][] w = new int[nRide][nCar];
//        for(int j = 0; j < nRide; j++)
//            for(int l = 0; l < nCar; l++){
//                Type1 car = input.type1List.get(l);
//                Type2 ride = input.type2List.get(j);
//                if((ride.tw_a >= car.end * 60)||(ride.tw_b <= car.startD * 60)){
//                    w[j][l] = 0;
//                }else if(ride.tw_a <= car.startD * 60){
//                    if(car.startD * 60 + input.travelMatrix[ride.origin][ride.destin] <= car.end * 60){
//                        w[j][l] = 1;
//                    }else{
//                        w[j][l] = 0;
//                    }
//                }else{
//                    if(ride.tw_a + input.travelMatrix[ride.origin][ride.destin] <= car.end * 60){
//                        w[j][l] = 1;
//                    }else{
//                        w[j][l] = 0;
//                    }
//                }
//            }
        List<Integer> vehicle = new ArrayList<Integer>();
        for(int l = 0; l < input.nType1; l++){
                vehicle.add(l);
        }
        //generate nodes
        List<Node> V0 = new ArrayList<Node>(); // depot start
        List<Node> V1 = new ArrayList<Node>(); // depot end
        List<Node> V2 = new ArrayList<Node>(); // ride
//        List<Node> V3 = new ArrayList<Node>();

        counter = 0;
        for(int l : vehicle){
            Type1 cr = input.type1List.get(l);
            Node node = new Node();
            node.index = counter;
            node.location = cr.origin;
            node.tw_a = cr.startD * 60;
            node.tw_b = cr.end * 60;
            node.idx_car = l;
            V0.add(node);
            counter++;
        }

        for(int l : vehicle){
            Type1 cr = input.type1List.get(l);
            Node node = new Node();
            node.index = counter;
            node.location = cr.destin;
            node.tw_a = cr.startD * 60;
            node.tw_b = cr.end * 60;
            node.idx_car = l;
            V1.add(node);
            counter++;
        }

        for(int j = 0; j < nRide; j++){
            Type2 rr = input.type2List.get(j);
            Node node = new Node();
            node.index = counter;
            node.location = rr.origin;
            node.tw_a = rr.tw_a;
            node.tw_b = rr.tw_b;
            node.idx_ride = j;
            V2.add(node);
            counter++;
        }

//        for(int j = 0; j < nRide; j++){
//            Type2 rr = input.type2List.get(j);
//            Node node = new Node();
//            node.index = counter;
//            node.location = rr.destin;
//            node.tw_a = rr.tw_a;
//            node.tw_b = 24 * 60;
//            node.idx_ride = j;
//            V3.add(node);
//            counter++;
//        }
        // generate Edge
        List<Edge2> E0 = new ArrayList<Edge2>(); // leave depot
        List<Edge2> E1 = new ArrayList<Edge2>(); // travel
        List<Edge2> E2 = new ArrayList<Edge2>(); // return depot
        //List<Edge2> E3 = new ArrayList<Edge2>();

        counter = 0;
        for(Node u : V0){
            for(Node v : V2){
                Edge2 edge = new Edge2();
                edge.node1 = u;
                edge.node2 = v;
                edge.index = counter;
                E0.add(edge);
                counter++;
            }
        }

        for(Node u : V2){
            for(Node v : V2){
                if(u.idx_ride == v.idx_ride){
                    continue;
                }else{
                    Edge2 edge = new Edge2();
                    edge.node1 = u;
                    edge.node2 = v;
                    edge.index = counter;
                    E1.add(edge);
                    counter++;
                }
            }
        }


        for(Node u : V2){
            for(Node v : V1){
                Edge2 edge = new Edge2();
                edge.node1 = u;
                edge.node2 = v;
                edge.index = counter;
                E2.add(edge);
                counter++;
            }
        }

        List<Node> V = new ArrayList<Node>();
        V.addAll(V0);
        V.addAll(V1);
        V.addAll(V2);
        List<Edge2> E = new ArrayList<Edge2>();
        E.addAll(E0);
        E.addAll(E1);
        E.addAll(E2);

        try{
            //Model
            String dirName = logDir + input.dirName;
            f = new File(dirName);
            if(!f.exists()){
                f.mkdir();
            }
            String logName = logDir + input.dirName+ input.fileName +"_gurobi_phase1_rel";
            f = new File(logName);
            f.delete();

            GRBEnv env = new GRBEnv(logName);
            GRBModel model = new GRBModel(env);
            model.set(GRB.DoubleParam.TimeLimit, 1800.0);
            GRBVar[] x = new GRBVar[nZone];
            for(int i = 0; i < nZone; i++){
                x[i] = model.addVar(0, nCars, 0, GRB.INTEGER, "x"+i);
            }
            //			System.out.println(nCars);
            GRBVar[] y = new GRBVar[edgeList.size()];
            for(int i = 0; i < edgeList.size(); i++){
                if(edgeList.get(i).type == 1){
                    y[i] = model.addVar(0, 1, 0, GRB.BINARY, "y"+i);
                }else {
                    y[i] = model.addVar(0, nCars, 0, GRB.INTEGER, "y"+i);
                }
            }

            GRBVar[][] alpha = new GRBVar[E.size()][vehicle.size()];
            for(int i = 0; i < E0.size(); i++){
                for(int l = 0; l < vehicle.size(); l++){
                    if(E0.get(i).node1.idx_car == vehicle.get(l))
                        alpha[i][l] = model.addVar(0, 1, 0, GRB.BINARY, "alpha_"+i+"_"+l);
                    else{
                        alpha[i][l] = model.addVar(0, 0, 0, GRB.BINARY, "alpha_"+i+"_"+l);
                    }
                }

            }
            for(int i = E0.size(); i < E.size(); i++){
                for(int l = 0; l < vehicle.size(); l++){
                    alpha[i][l] = model.addVar(0, 1, 0, GRB.BINARY, "alpha_"+i+"_"+l);
                }

            }
            GRBVar[] beta = new GRBVar[nRide];
            for(int j = 0; j < nRide; j++){
                beta[j] = model.addVar(0, 1, 0, GRB.BINARY, "beta"+j);
            }

            GRBVar[] gamma = new GRBVar[V.size()];
            for(int i = 0; i < V.size(); i++){
                Node node = V.get(i);
                gamma[i] = model.addVar(node.tw_a, node.tw_b, 0, GRB.CONTINUOUS, "gamma"+i);
            }
            // Integrate new variables

            model.update();


            GRBLinExpr expr = new GRBLinExpr();
//            for(Edge e : tripEdge){
//                expr.addTerm(1, y[e.idx]);
//            }
            for(Edge e : edgeList){
                expr.addTerm(e.cost, y[e.idx]);
            }
            for(int i = 0; i < nRide; i++){
                Type2 rr = input.type2List.get(i);
                expr.addTerm(ratePerHourRide / 60.0 * input.travelMatrix[rr.origin][rr.destin], beta[i]);
                //expr.addTerm(1, beta[i]);

            }
//            for(int j = 0; j < input.type2List.size(); j++){
//                for(int l = 0; l < input.type1List.size(); l++){
//                    expr.addTerm(1, z[j][l]);
//                }
//            }
            model.setObjective(expr, GRB.MAXIMIZE);

            expr.clear();
            for(int i = 0 ; i < nZone; i++){
                expr.addTerm(1, x[i]);
            }
            model.addConstr(expr, GRB.EQUAL, nCars, "initial location");

            for(int i = 0; i < nZone; i++){
                expr.clear();
                for(int edge : adj.get(i).get(0)){
                    expr.addTerm(1, y[edge]);
                }
                expr.addTerm(-1, x[i]);
                model.addConstr(expr, GRB.EQUAL, 0, "initial flow "+i);
            }

            for(int t = 1; t < nPeriod; t++){
                for(int i = 0; i < nZone; i++){
                    expr.clear();
                    for(int edge : adj.get(i).get(t)){
                        expr.addTerm(1, y[edge]);
                    }
                    for(int edge: inj.get(i).get(t)){
                        expr.addTerm(-1, y[edge]);

                    }
                    model.addConstr(expr, GRB.EQUAL, 0, "flow balance"+"-"+t);
                }
            }


            for(int j = 0; j < nRide; j++){
                expr.clear();
                for(int l = 0; l < vehicle.size(); l++){
                    for(Edge2 e : E0){
                        if(e.node2.idx_ride == j){
                            expr.addTerm(1, alpha[e.index][l]);
                        }
                    }
                    for(Edge2 e : E1){
                        if(e.node2.idx_ride == j){
                            expr.addTerm(1, alpha[e.index][l]);
                        }
                    }
                }
                model.addConstr(expr, GRB.EQUAL, beta[j], "beta"+j);
            }

            // depart constraints
            for(int l = 0; l < vehicle.size(); l++){
                expr.clear();
                for(Edge2 e : E0){
                    if(e.node1.idx_car == vehicle.get(l)){
                        expr.addTerm(1, alpha[e.index][l]);
                    }
                }
                for(int i = 0; i < edgeList.size(); i++) {
                    if (edgeList.get(i).idx == l) {
                       expr.addTerm(-1,y[i]);
                    }
                }
                model.addConstr(expr, GRB.LESS_EQUAL, 0, "initial"+l);
            }


            // end constraints
            for(int l = 0; l < vehicle.size(); l++){
                expr.clear();
                for(Edge2 e : E2){
                    if(e.node2.idx_car == vehicle.get(l)){
                        expr.addTerm(1, alpha[e.index][l]);
                    }
                }
                for(Edge2 e : E0){
                    if(e.node1.idx_car == vehicle.get(l)){
                        expr.addTerm(-1, alpha[e.index][l]);
                    }
                }
                model.addConstr(expr, GRB.EQUAL, 0, "end"+l);
            }

            //flow balance
            for(int l = 0; l < vehicle.size(); l++){
                for(Node v : V2){
                    expr.clear();
                    for(Edge2 e : E){
                        if(e.node2.idx_ride == v.idx_ride){
                            expr.addTerm(1, alpha[e.index][l]);
                        }else if(e.node1.idx_ride == v.idx_ride){
                            expr.addTerm(-1, alpha[e.index][l]);
                        }
                    }
                    model.addConstr(expr, GRB.EQUAL, 0, "flowbalance"+v.index+"-"+l);
                }
            }



            for(Edge2 e: E0){
                Node u = e.node1;
                Node v = e.node2;
                expr.clear();
                expr.addTerm(1,gamma[u.index]);
                expr.addTerm(-1, gamma[v.index]);
                for(int l = 0; l < vehicle.size(); l++){
                    if(vehicle.get(l) == u.idx_car){
                        expr.addTerm(12*60, alpha[e.index][l]);
                    }
                }
                model.addConstr(expr, GRB.LESS_EQUAL, 12*60 - input.travelMatrix[u.location][v.location], "time arrival"+e.index);
            }

            for(Edge2 e: E1){
                Node u = e.node1;
                Node v = e.node2;
                Type2 temp = input.type2List.get(u.idx_ride);
                expr.clear();
                expr.addTerm(1,gamma[u.index]);
                expr.addTerm(-1, gamma[v.index]);
                for(int l = 0; l < vehicle.size(); l++){
                    expr.addTerm(12*60, alpha[e.index][l]);
                }
                model.addConstr(expr, GRB.LESS_EQUAL, 12*60 - input.travelMatrix[temp.origin][temp.destin] -
                        input.travelMatrix[temp.destin][v.location], "time arrival"+e.index);
            }

            for(Edge2 e: E2){
                Node u = e.node1;
                Node v = e.node2;
                Type2 temp = input.type2List.get(u.idx_ride);
                expr.clear();
                expr.addTerm(1,gamma[u.index]);
                expr.addTerm(-1, gamma[v.index]);
                for(int l = 0; l < vehicle.size(); l++){
                    //if (vehicle.get(l) == v.idx_car) {
                        expr.addTerm(12 * 60, alpha[e.index][l]);
                    //}
                }
                model.addConstr(expr, GRB.LESS_EQUAL, 12*60 - input.travelMatrix[temp.origin][temp.destin] -
                        input.travelMatrix[temp.destin][v.location], "time arrival"+e.index);
            }
//            for(int i = 0; i < nZone; i++){
//                expr.clear();
//                for(int edge : inj.get(i).get(nPeriod)){
//                    expr.addTerm(1, y[edge]);
//                }
//                expr.addTerm(-1, x[i]);
//                model.addConstr(expr, GRB.EQUAL, 0, "terminal flow "+i);
//            }

//            for(int l = 0; l < nCar; l++){
//                Type1 car = input.type1List.get(l);
//                expr.clear();
//                for(int j = 0; j < nRide; j++){
//                    Type2 ride = input.type2List.get(j);
//                    expr.addTerm(input.travelMatrix[ride.origin][ride.destin], z[j][l]);
//                }
//                expr.addTerm(-60 * (car.end - car.startD), y[l]);
//                model.addConstr(expr, GRB.LESS_EQUAL, 0, "matching"+l);
//            }
//
//            for(int j = 0; j < nRide; j++){
//                expr.clear();
//                for(int l = 0; l < nCar; l++){
//                    expr.addTerm(1, z[j][l]);
//                }
//                model.addConstr(expr, GRB.LESS_EQUAL, 1, "matched"+j);
//            }

            model.update();
            model.optimize();

            if(model.get(GRB.IntAttr.Status) != GRB.Status.OPTIMAL){
                System.err.println("Error: model is infeasible");
                return;
            }
            double lowerbound = model.get(GRB.DoubleAttr.ObjBound);
            double upperbound = model.get(GRB.DoubleAttr.ObjVal);
            result.gap_phaseI = 1- lowerbound/upperbound;


            double[] solution_x = model.get(GRB.DoubleAttr.X, x);
//            System.out.println("Solution x:"+Arrays.toString(solution_x));
            result.rawSolution_x = solution_x;
            double[] solution_y = model.get(GRB.DoubleAttr.X, y);
            result.rawSolution_y = solution_y;

            double[][] idle = new double[nZone][nPeriod];
            for(int i = 0; i < nZone; i++){
                for(int t = 0; t < nPeriod; t++){
                    for(int idx : adj.get(i).get(t)){
                        if( edgeList.get(idx).type ==  2){
                            //System.out.println("["+i+","+t+"]: "+solution_y[idx]);
                            idle[i][t] = solution_y[idx];
                        }
                    }
                }
            }
            result.idleness = idle;


            double[][][] relocation = new double[nPeriod][nZone][nZone];
            for(int i = 0; i < nZone; i++){
                for(int t = 0; t < nPeriod; t++){
                    for(int idx : adj.get(i).get(t)){
                        if(edgeList.get(idx).type == 3){
                            relocation[t][i][edgeList.get(idx).loc2] = solution_y[idx];
                        }
                    }
                }
            }
            result.relocation = relocation;


            int count = 0;
            for(int l = 0 ; l < nCar; l++){
                if(solution_y[l] >= 1-1E-6){
                    input.type1List.get(l).approved = true;
                    count += 1;
                }
            }
            result.fulfillmentRate_Car = ((double) count) / nCar;
//            System.out.println("Solution y:"+Arrays.toString(solution_y));
            double runtime = model.get(GRB.DoubleAttr.Runtime);
            result.runtime_phaseI = (double) Math.round(runtime * 1000d) / 1000d;
            result.outputPhaseI();
            model.dispose();
            env.dispose();

        } catch (GRBException e) {
            System.out.println("Error code: " + e.getErrorCode() + ". " +
                    e.getMessage());
        }
    }

    public void solvePhaseII_deterministic(){
        List<Integer> vehicle = new ArrayList<Integer>();
        for(int l = 0; l < input.nType1; l++){
            if(input.type1List.get(l).approved){
                vehicle.add(l);
            }
        }
        //generate nodes
        List<Node> V0 = new ArrayList<Node>(); // depot start
        List<Node> V1 = new ArrayList<Node>(); // depot end
        List<Node> V2 = new ArrayList<Node>(); // ride
//        List<Node> V3 = new ArrayList<Node>();

        int nCar = input.nType1;
        int nRide = input.nType2;

        int counter = 0;
        for(int l : vehicle){
            Type1 cr = input.type1List.get(l);
            Node node = new Node();
            node.index = counter;
            node.location = cr.origin;
            node.tw_a = cr.startD * 60;
            node.tw_b = cr.end * 60;
            node.idx_car = l;
            V0.add(node);
            counter++;
        }

        for(int l : vehicle){
            Type1 cr = input.type1List.get(l);
            Node node = new Node();
            node.index = counter;
            node.location = cr.destin;
            node.tw_a = cr.startD * 60;
            node.tw_b = cr.end * 60;
            node.idx_car = l;
            V1.add(node);
            counter++;
        }

        for(int j = 0; j < nRide; j++){
            Type2 rr = input.type2List.get(j);
            Node node = new Node();
            node.index = counter;
            node.location = rr.origin;
            node.tw_a = rr.tw_a;
            node.tw_b = rr.tw_b;
            node.idx_ride = j;
            V2.add(node);
            counter++;
        }

//        for(int j = 0; j < nRide; j++){
//            Type2 rr = input.type2List.get(j);
//            Node node = new Node();
//            node.index = counter;
//            node.location = rr.destin;
//            node.tw_a = rr.tw_a;
//            node.tw_b = 24 * 60;
//            node.idx_ride = j;
//            V3.add(node);
//            counter++;
//        }
        // generate Edge
        List<Edge2> E0 = new ArrayList<Edge2>(); // leave depot
        List<Edge2> E1 = new ArrayList<Edge2>(); // travel
        List<Edge2> E2 = new ArrayList<Edge2>(); // return depot
        //List<Edge2> E3 = new ArrayList<Edge2>();

        counter = 0;
        for(Node u : V0){
            for(Node v : V2){
                Edge2 edge = new Edge2();
                edge.node1 = u;
                edge.node2 = v;
                edge.index = counter;
                E0.add(edge);
                counter++;
            }
        }

        for(Node u : V2){
            for(Node v : V2){
                if(u.idx_ride == v.idx_ride){
                    continue;
                }else{
                    Edge2 edge = new Edge2();
                    edge.node1 = u;
                    edge.node2 = v;
                    edge.index = counter;
                    E1.add(edge);
                    counter++;
                }
            }
        }


        for(Node u : V2){
            for(Node v : V1){
                Edge2 edge = new Edge2();
                edge.node1 = u;
                edge.node2 = v;
                edge.index = counter;
                E2.add(edge);
                counter++;
            }
        }

        List<Node> V = new ArrayList<Node>();
        V.addAll(V0);
        V.addAll(V1);
        V.addAll(V2);
        List<Edge2> E = new ArrayList<Edge2>();
        E.addAll(E0);
        E.addAll(E1);
        E.addAll(E2);


        try {
            String dirName = logDir + input.dirName;
            f = new File(dirName);
            if(!f.exists()){
                f.mkdir();
            }
            String logName = logDir + input.dirName+input.fileName +"_gurobi_phase2";
            f = new File(logName);
            f.delete();
            GRBEnv env = new GRBEnv(logName);
            GRBModel  model = new GRBModel(env);
            model.set(GRB.DoubleParam.TimeLimit, 10800.0);


            // Create variables
            GRBVar[][] alpha = new GRBVar[E.size()][vehicle.size()];
            for(int i = 0; i < E0.size(); i++){
                for(int l = 0; l < vehicle.size(); l++){
                    if(E0.get(i).node1.idx_car == vehicle.get(l))
                        alpha[i][l] = model.addVar(0, 1, 0, GRB.BINARY, "alpha_"+i+"_"+l);
                    else{
                        alpha[i][l] = model.addVar(0, 0, 0, GRB.BINARY, "alpha_"+i+"_"+l);
                    }
                }

            }
            for(int i = E0.size(); i < E.size(); i++){
                for(int l = 0; l < vehicle.size(); l++){
                    alpha[i][l] = model.addVar(0, 1, 0, GRB.BINARY, "alpha_"+i+"_"+l);
                }

            }
            GRBVar[] beta = new GRBVar[nRide];
            for(int j = 0; j < nRide; j++){
                beta[j] = model.addVar(0, 1, 0, GRB.BINARY, "beta"+j);
            }

            GRBVar[] gamma = new GRBVar[V.size()];
            for(int i = 0; i < V.size(); i++){
                Node node = V.get(i);
                gamma[i] = model.addVar(node.tw_a, node.tw_b, 0, GRB.CONTINUOUS, "gamma"+i);
            }
            // Integrate new variables

            model.update();

            // Set objective: max visited node
            GRBLinExpr expr = new GRBLinExpr();
            for(int i = 0; i < nRide; i++){
                Type2 rr = input.type2List.get(i);
                expr.addTerm(ratePerHourRide / 60.0 * input.travelMatrix[rr.origin][rr.destin], beta[i]);
                //expr.addTerm(1, beta[i]);

            }

            model.setObjective(expr, GRB.MAXIMIZE);


            for(int j = 0; j < nRide; j++){
                expr.clear();
                for(int l = 0; l < vehicle.size(); l++){
                    for(Edge2 e : E0){
                        if(e.node2.idx_ride == j){
                            expr.addTerm(1, alpha[e.index][l]);
                        }
                    }
                    for(Edge2 e : E1){
                        if(e.node2.idx_ride == j){
                            expr.addTerm(1, alpha[e.index][l]);
                        }
                    }
                }
                model.addConstr(expr, GRB.EQUAL, beta[j], "beta"+j);
            }

            // depart constraints
            for(int l = 0; l < vehicle.size(); l++){
                expr.clear();
                for(Edge2 e : E0){
                    if(e.node1.idx_car == vehicle.get(l)){
                        expr.addTerm(1, alpha[e.index][l]);
                    }
                }
                model.addConstr(expr, GRB.LESS_EQUAL, 1, "initial"+l);
            }


            // end constraints
            for(int l = 0; l < vehicle.size(); l++){
                expr.clear();
                for(Edge2 e : E2){
                    if(e.node2.idx_car == vehicle.get(l)){
                        expr.addTerm(1, alpha[e.index][l]);
                    }
                }
                for(Edge2 e : E0){
                    if(e.node1.idx_car == vehicle.get(l)){
                        expr.addTerm(-1, alpha[e.index][l]);
                    }
                }
                model.addConstr(expr, GRB.EQUAL, 0, "end"+l);
            }

            //flow balance
            for(int l = 0; l < vehicle.size(); l++){
                for(Node v : V2){
                    expr.clear();
                    for(Edge2 e : E){
                        if(e.node2.idx_ride == v.idx_ride){
                            expr.addTerm(1, alpha[e.index][l]);
                        }else if(e.node1.idx_ride == v.idx_ride){
                            expr.addTerm(-1, alpha[e.index][l]);
                        }
                    }
                    model.addConstr(expr, GRB.EQUAL, 0, "flowbalance"+v.index+"-"+l);
                }
            }

            //subtour elimination and time arrival
            for(Edge2 e: E0){
                Node u = e.node1;
                Node v = e.node2;
                expr.clear();
                expr.addTerm(1,gamma[u.index]);
                expr.addTerm(-1, gamma[v.index]);
                for(int l = 0; l < vehicle.size(); l++){
                    if(vehicle.get(l) == u.idx_car){
                        expr.addTerm(12*60, alpha[e.index][l]);
                    }
                }
                model.addConstr(expr, GRB.LESS_EQUAL, 12*60 - input.travelMatrix[u.location][v.location], "time arrival"+e.index);
            }

            for(Edge2 e: E1){
                Node u = e.node1;
                Node v = e.node2;
                Type2 temp = input.type2List.get(u.idx_ride);
                expr.clear();
                expr.addTerm(1,gamma[u.index]);
                expr.addTerm(-1, gamma[v.index]);
                for(int l = 0; l < vehicle.size(); l++){
                    expr.addTerm(12*60, alpha[e.index][l]);
                }
                model.addConstr(expr, GRB.LESS_EQUAL, 12*60 - input.travelMatrix[temp.origin][temp.destin] -
                        input.travelMatrix[temp.destin][v.location], "time arrival"+e.index);
            }

            for(Edge2 e: E2){
                Node u = e.node1;
                Node v = e.node2;
                Type2 temp = input.type2List.get(u.idx_ride);
                expr.clear();
                expr.addTerm(1,gamma[u.index]);
                expr.addTerm(-1, gamma[v.index]);
                for(int l = 0; l < vehicle.size(); l++){
                    if(vehicle.get(l) == v.idx_car) {
                        expr.addTerm(12 * 60, alpha[e.index][l]);
                    }
                }
                model.addConstr(expr, GRB.LESS_EQUAL, 12*60 - input.travelMatrix[temp.origin][temp.destin] -
                        input.travelMatrix[temp.destin][v.location], "time arrival"+e.index);
            }


            // Optimize model
//            // Set the TuneResults parameter to 1
//            model.set(GRB.IntParam.TuneResults, 1);
//
//            // Tune the model
//            model.tune();
//
//            // Get the number of tuning results
//            int resultcount = model.get(GRB.IntAttr.TuneResultCount);
//
//            if (resultcount > 0) {
//
//                // Load the tuned parameters into the model's environment
//                model.getTuneResult(0);
//
//                // Write the tuned parameters to a file
//                model.write("data/tune/"+input.fileName+"deterministic_tune.prm");
//
//                // Solve the model using the tuned parameters
//                model.optimize();
//            }

//            model.set(GRB.IntParam.GomoryPasses, 5);
//            model.set(GRB.IntParam.MIPFocus, 2);
            model.optimize();
            System.out.println(alpha.length);
            double runtime = model.get(GRB.DoubleAttr.Runtime);
            result.runtime_phaseII_deterministic = (double) Math.round(runtime * 1000d) / 1000d;

            double[][] solution_alpha = model.get(GRB.DoubleAttr.X, alpha);
            //System.out.println("Solution alpha:"+Arrays.toString(solution_alpha[799]));
            result.rawSolution_alpha = solution_alpha;


            double[] solution_beta = model.get(GRB.DoubleAttr.X, beta);
            System.out.println("Solution beta:"+Arrays.toString(solution_beta));
            result.rawSolution_beta = solution_beta;
            int count = 0;
            for(int i = 0; i < nRide; i++){
                if(solution_beta[i]  > 0.5){
                    count += 1;
                }
            }
            result.fulfillmentRate_d_Ride = ((double) count )/ nRide;

            double[] solution_gamma = model.get(GRB.DoubleAttr.X, gamma);
            result.rawSolution_gamma = solution_gamma;
            System.out.println("Solution gamma:"+Arrays.toString(solution_gamma));

            //			for(Edge2 e : E2){
            //				for(int l = 0; l < vehicle.size(); l++){
            //					if(solution_alpha[e.index][l] > 0.5){
            //						System.out.println("Ride "+e.node2.idx_ride+" is served by "+e.node1.idx_car+"("+l+")" );
            //					}
            //				}
            //			}
            //
            ArrayList<ArrayList<Integer>> arrangement = new ArrayList<>();
            for(int i = 0; i < nCar; i++){
                arrangement.add(new ArrayList<>());
            }


            for(Edge2 e : E0){
                for(int l = 0; l < vehicle.size(); l++){
                    if(solution_alpha[e.index][l] > 0.5){
                        //System.out.println("Ride "+e.node2.idx_ride+" is served by "+e.node1.idx_car+"("+l+")" );
                        arrangement.get(vehicle.get(l)).add(e.node2.idx_ride);
                    }
                }
            }
            for(int l = 0; l < vehicle.size(); l++){
                boolean flag = true;
                while (flag){
                    flag = false;
                    if(arrangement.get(vehicle.get(l)).size() < 1)
                        continue;
                    int temp = arrangement.get(vehicle.get(l)).get(arrangement.get(vehicle.get(l)).size()-1);
                    for(Edge2 e : E1){
                        if(solution_alpha[e.index][l] > 0.5 && e.node1.idx_ride == temp){
                            //System.out.println("Ride "+e.node2.idx_ride+" is served by "+e.node1.idx_car+"("+l+")" );
                            arrangement.get(vehicle.get(l)).add(e.node2.idx_ride);
                            flag = true;
                            System.out.println(vehicle.get(l)+","+temp);

                        }
                    }
                }

            }

            for(int i = 0; i < nCar; i++){
                System.out.println("arrangement for car "+i+": "+arrangement.get(i).toString());
            }

            double[] newArray = Arrays.copyOfRange(solution_gamma, vehicle.size() * 2, solution_gamma.length);

            //System.out.println(Arrays.toString(newArray));
            result.assignment = arrangement;

//            System.out.println("Obj: " + model.get(GRB.DoubleAttr.ObjVal));
            double upperbound = model.get(GRB.DoubleAttr.ObjBound);
            double lowerbound = model.get(GRB.DoubleAttr.ObjVal);
            result.gap_phaseII = 1- lowerbound/upperbound;
            // Dispose of model and environment
            result.outputPhaseII();
            model.dispose();
            env.dispose();

        } catch (GRBException e) {
            System.out.println("Error code: " + e.getErrorCode() + ". " +
                    e.getMessage());
        }
    }

    public void solvePhaseII_stochastic(int sample, int index){
        input.loadStochastic(sample, index);
        List<Integer> vehicle = new ArrayList<Integer>();
        for(int l = 0; l < input.nType1; l++){
            if(input.type1List.get(l).approved){
                vehicle.add(l);
            }
        }
        //generate nodes
        List<Node> V0 = new ArrayList<Node>(); // depot start
        List<Node> V1 = new ArrayList<Node>(); // depot end
        List<Node> V2 = new ArrayList<Node>(); // ride
//        List<Node> V3 = new ArrayList<Node>();

        int nCar = input.nType1;
        int nRide = input.nType2;

        int counter = 0;
        for(int l : vehicle){
            Type1 cr = input.type1List.get(l);
            Node node = new Node();
            node.index = counter;
            node.location = cr.origin;
            node.tw_a = cr.startD * 60;
            node.tw_b = cr.end * 60;
            node.idx_car = l;
            V0.add(node);
            counter++;
        }

        for(int l : vehicle){
            Type1 cr = input.type1List.get(l);
            Node node = new Node();
            node.index = counter;
            node.location = cr.destin;
            node.tw_a = cr.startD * 60;
            node.tw_b = cr.end * 60;
            node.idx_car = l;
            V1.add(node);
            counter++;
        }

        for(int j = 0; j < nRide; j++){
            Type2 rr = input.type2List.get(j);
            Node node = new Node();
            node.index = counter;
            node.location = rr.origin;
            node.tw_a = rr.tw_a;
            node.tw_b = rr.tw_b;
            node.idx_ride = j;
            V2.add(node);
            counter++;
        }

//        for(int j = 0; j < nRide; j++){
//            Type2 rr = input.type2List.get(j);
//            Node node = new Node();
//            node.index = counter;
//            node.location = rr.destin;
//            node.tw_a = rr.tw_a;
//            node.tw_b = 24 * 60;
//            node.idx_ride = j;
//            V3.add(node);
//            counter++;
//        }
        // generate Edge
        List<Edge2> E0 = new ArrayList<Edge2>(); // leave depot
        List<Edge2> E1 = new ArrayList<Edge2>(); // travel
        List<Edge2> E2 = new ArrayList<Edge2>(); // return depot
        //List<Edge2> E3 = new ArrayList<Edge2>();

        counter = 0;
        for(Node u : V0){
            for(Node v : V2){
                Edge2 edge = new Edge2();
                edge.node1 = u;
                edge.node2 = v;
                edge.index = counter;
                E0.add(edge);
                counter++;
            }
        }

        for(Node u : V2){
            for(Node v : V2){
                if(u.idx_ride == v.idx_ride){
                    continue;
                }else{
                    Edge2 edge = new Edge2();
                    edge.node1 = u;
                    edge.node2 = v;
                    edge.index = counter;
                    E1.add(edge);
                    counter++;
                }
            }
        }


        for(Node u : V2){
            for(Node v : V1){
                Edge2 edge = new Edge2();
                edge.node1 = u;
                edge.node2 = v;
                edge.index = counter;
                E2.add(edge);
                counter++;
            }
        }

        List<Node> V = new ArrayList<Node>();
        V.addAll(V0);
        V.addAll(V1);
        V.addAll(V2);
        List<Edge2> E = new ArrayList<Edge2>();
        E.addAll(E0);
        E.addAll(E1);
        E.addAll(E2);


        try {
            String dirName = logDir + input.dirName;
            f = new File(dirName);
            if(!f.exists()){
                f.mkdir();
            }
            String logName = logDir + input.dirName + input.fileName +"_gurobi_"+sample+"_"+index;
            f = new File(logName);
            f.delete();
            GRBEnv env = new GRBEnv(logName);
            GRBModel  model = new GRBModel(env);
            model.set(GRB.DoubleParam.TimeLimit, 10800.0);


            // Create variables
            GRBVar[][] alpha = new GRBVar[E.size()][vehicle.size()];
            for(int i = 0; i < E0.size(); i++){
                for(int l = 0; l < vehicle.size(); l++){
                    if(E0.get(i).node1.idx_car == vehicle.get(l))
                        alpha[i][l] = model.addVar(0, 1, 0, GRB.BINARY, "alpha_"+i+"_"+l);
                    else{
                        alpha[i][l] = model.addVar(0, 0, 0, GRB.BINARY, "alpha_"+i+"_"+l);
                    }
                }

            }
            for(int i = E0.size(); i < E.size(); i++){
                for(int l = 0; l < vehicle.size(); l++){
                    alpha[i][l] = model.addVar(0, 1, 0, GRB.BINARY, "alpha_"+i+"_"+l);
                }

            }
            GRBVar[] beta = new GRBVar[nRide];
            for(int j = 0; j < nRide; j++){
                beta[j] = model.addVar(0, 1, 0, GRB.BINARY, "beta"+j);
            }

            GRBVar[] gamma = new GRBVar[V.size()];
            for(int i = 0; i < V.size(); i++){
                Node node = V.get(i);
                gamma[i] = model.addVar(node.tw_a, node.tw_b, 0, GRB.CONTINUOUS, "gamma"+i);
            }
            GRBVar[][] wait = new GRBVar[nRide][sample];
            for(int j = 0; j < nRide; j++){
                for(int s = 0; s < sample; s++){
                    wait[j][s] = model.addVar(0, 12*60, 0, GRB.CONTINUOUS, "wait"+j+"_"+s);
                }
            }
//            GRBVar[][] idle = new GRBVar[nRide][sample];
//            for(int j = 0; j < nRide; j++){
//                for(int s = 0; s < sample; s++){
//                    idle[j][s] = model.addVar(0, 12 *60, 0, GRB.CONTINUOUS, "idle"+j+"_"+s);
//                }
//            }

            GRBVar[][] overtime = new GRBVar[nCar][sample];
            for(int l = 0; l < nCar; l++){
                for(int s = 0 ; s < sample; s++){
                    if(vehicle.contains(l)){
                        overtime[l][s] = model.addVar(0, GRB.INFINITY, 0, GRB.CONTINUOUS, "overtime"+l+"_"+s);
                    }else{
                        overtime[l][s] = model.addVar(0, 0, 0, GRB.CONTINUOUS, "overtime"+l+"_"+s);
                    }
                }
            }
            model.update();

            // Set objective: max visited node
            GRBLinExpr expr = new GRBLinExpr();
            for(int i = 0; i < nRide; i++){
                Type2 rr = input.type2List.get(i);
                expr.addTerm(ratePerHourRide / 60.0 * input.travelMatrix[rr.origin][rr.destin], beta[i]);
                //expr.addTerm(1, beta[i]);

            }
            for(int s = 0; s < sample; s++){
                for(int l : vehicle){
                    expr.addTerm(-0.1/sample, overtime[l][s]);
                }
                for(int j = 0; j < nRide; j++){
                    expr.addTerm(-0.1/sample, wait[j][s]);
                }
            }

            model.setObjective(expr, GRB.MAXIMIZE);


            for(int j = 0; j < nRide; j++){
                expr.clear();
                for(int l = 0; l < vehicle.size(); l++){
                    for(Edge2 e : E0){
                        if(e.node2.idx_ride == j){
                            expr.addTerm(1, alpha[e.index][l]);
                        }
                    }
                    for(Edge2 e : E1){
                        if(e.node2.idx_ride == j){
                            expr.addTerm(1, alpha[e.index][l]);
                        }
                    }
                }
                model.addConstr(expr, GRB.EQUAL, beta[j], "beta"+j);
            }

            // depart constraints
            for(int l = 0; l < vehicle.size(); l++){
                expr.clear();
                for(Edge2 e : E0){
                    if(e.node1.idx_car == vehicle.get(l)){
                        expr.addTerm(1, alpha[e.index][l]);
                    }
                }
                model.addConstr(expr, GRB.LESS_EQUAL, 1, "initial"+l);
            }


            // end constraints
            for(int l = 0; l < vehicle.size(); l++){
                expr.clear();
                for(Edge2 e : E2){
                    if(e.node2.idx_car == vehicle.get(l)){
                        expr.addTerm(1, alpha[e.index][l]);
                    }
                }
                for(Edge2 e : E0){
                    if(e.node1.idx_car == vehicle.get(l)){
                        expr.addTerm(-1, alpha[e.index][l]);
                    }
                }
                model.addConstr(expr, GRB.EQUAL, 0, "end"+l);
            }

            //flow balance
            for(int l = 0; l < vehicle.size(); l++){
                for(Node v : V2){
                    expr.clear();
                    for(Edge2 e : E){
                        if(e.node2.idx_ride == v.idx_ride){
                            expr.addTerm(1, alpha[e.index][l]);
                        }else if(e.node1.idx_ride == v.idx_ride){
                            expr.addTerm(-1, alpha[e.index][l]);
                        }
                    }
                    model.addConstr(expr, GRB.EQUAL, 0, "flowbalance"+v.index+"-"+l);
                }
            }

            //subtour elimination and time arrival
            for(Edge2 e: E0){
                Node u = e.node1;
                Node v = e.node2;
                expr.clear();
                expr.addTerm(1,gamma[u.index]);
                expr.addTerm(-1, gamma[v.index]);
                for(int l = 0; l < vehicle.size(); l++){
                    if(vehicle.get(l) == u.idx_car){
                        expr.addTerm(12*60, alpha[e.index][l]);
                    }
                }
                model.addConstr(expr, GRB.LESS_EQUAL, 12*60 - input.travelMatrix[u.location][v.location], "time arrival"+e.index);
            }

            for(Edge2 e: E1){
                Node u = e.node1;
                Node v = e.node2;
                Type2 temp = input.type2List.get(u.idx_ride);
                expr.clear();
                expr.addTerm(1,gamma[u.index]);
                expr.addTerm(-1, gamma[v.index]);
                for(int l = 0; l < vehicle.size(); l++){
                    expr.addTerm(12*60, alpha[e.index][l]);
                }
                model.addConstr(expr, GRB.LESS_EQUAL, 12*60 - input.travelMatrix[temp.origin][temp.destin] -
                        input.travelMatrix[temp.destin][v.location], "time arrival"+e.index);
            }

            for(Edge2 e: E2){
                Node u = e.node1;
                Node v = e.node2;
                Type2 temp = input.type2List.get(u.idx_ride);
                expr.clear();
                expr.addTerm(1,gamma[u.index]);
                expr.addTerm(-1, gamma[v.index]);
                for(int l = 0; l < vehicle.size(); l++){
                    expr.addTerm(12*60, alpha[e.index][l]);
                }
                model.addConstr(expr, GRB.LESS_EQUAL, 12*60 - input.travelMatrix[temp.origin][temp.destin] -
                        input.travelMatrix[temp.destin][v.location], "time arrival"+e.index);
            }

            // compute wait/idle time
            for(int s = 0; s < sample; s++){
                for(Edge2 e : E0){
                    Node u = e.node1;
                    Node v = e.node2;
                    expr.clear();
                    expr.addTerm(1,gamma[u.index]);
                    expr.addTerm(-1, gamma[v.index]);
                    expr.addTerm(-1, wait[v.idx_ride][s]);
                    for(int l = 0; l < vehicle.size(); l++){
                        if(vehicle.get(l) == u.idx_car){
                            expr.addTerm(12*60, alpha[e.index][l]);
                        }
                    }
                    //					System.out.println(Arrays.toString(ins.distance_s[9][0]));
                    model.addConstr(expr, GRB.LESS_EQUAL, 12*60 - input.stochasticTravelMatrix[s][u.location][v.location], "time arrival"+e.index);

                }
            }

            for(int s = 0; s < sample; s++){
                for(Edge2 e: E1){
                    Node u = e.node1;
                    Node v = e.node2;
                    Type2 temp = input.type2List.get(u.idx_ride);
                    expr.clear();
                    expr.addTerm(1,gamma[u.index]);
                    expr.addTerm(1, wait[u.idx_ride][s]);
                    //expr.addTerm(-1, idle[o.idx_ride][s]);
                    expr.addTerm(-1, gamma[v.index]);
                    expr.addTerm(-1, wait[v.idx_ride][s]);
                    //expr.addTerm(1, idle[v.idx_ride][s]);
                    for(int l = 0; l < vehicle.size(); l++){
                        expr.addTerm(12*60, alpha[e.index][l]);
                    }
                    model.addConstr(expr, GRB.LESS_EQUAL, 12*60 - input.stochasticTravelMatrix[s][temp.origin][temp.destin] - input.stochasticTravelMatrix[s][temp.destin][v.location], "time arrival"+e.index);
                }
            }



            for(int s = 0; s < sample; s++){
                for(Edge2 e: E2){
                    Node u = e.node1;
                    Node v = e.node2;
                    Type2 temp = input.type2List.get(u.idx_ride);
                    expr.clear();
                    expr.addTerm(1,gamma[u.index]);
                    expr.addTerm(1, wait[u.idx_ride][s]);
                    expr.addTerm(-1, gamma[v.index]);
                    expr.addTerm(-1, overtime[v.idx_car][s]);
                    for(int l = 0; l < vehicle.size(); l++){
                        if (vehicle.get(l) == u.idx_car) {

                            expr.addTerm(12 * 60, alpha[e.index][l]);
                        }
                    }
                    model.addConstr(expr, GRB.LESS_EQUAL, 12*60 - input.stochasticTravelMatrix[s][temp.origin][temp.destin] - input.stochasticTravelMatrix[s][temp.destin][v.location], "time arrival"+e.index);
                }
            }

//            // Set the TuneResults parameter to 1
//            model.set(GRB.IntParam.TuneResults, 1);
//
//            // Tune the model
//            model.tune();
//
//            // Get the number of tuning results
//            int resultcount = model.get(GRB.IntAttr.TuneResultCount);
//
//            if (resultcount > 0) {
//
//                // Load the tuned parameters into the model's environment
//                model.getTuneResult(0);
//
//                // Write the tuned parameters to a file
//                model.write("data/tune/"+input.fileName+"stochastic_tune.prm");
//
//                // Solve the model using the tuned parameters
//                model.optimize();
//            }
//            model.set(GRB.DoubleParam.Heuristics, 0);
            model.optimize();

            System.out.println(alpha.length);
            double runtime = model.get(GRB.DoubleAttr.Runtime);
            result.runtime_phaseII_stochastic = (double) Math.round(runtime * 1000d) / 1000d;

            double[][] solution_alpha = model.get(GRB.DoubleAttr.X, alpha);
            //System.out.println("Solution alpha:"+Arrays.toString(solution_alpha[799]));
            result.rawSolution_s_alpha = solution_alpha;


            double[] solution_beta = model.get(GRB.DoubleAttr.X, beta);
            System.out.println("Solution beta:"+Arrays.toString(solution_beta));
            result.rawSolution_s_beta = solution_beta;
            int count = 0;
            for(int i = 0; i < nRide; i++){
                if(solution_beta[i]  > 0.5){
                    count += 1;
                }
            }
            result.fulfillmentRate_s_Ride = ((double) count )/ nRide;

            double[][] waitTime = model.get(GRB.DoubleAttr.X, wait);
            result.rawSolution_wait = waitTime;


            result.rawSolution_overtime = model.get(GRB.DoubleAttr.X, overtime);


            double[] solution_gamma = model.get(GRB.DoubleAttr.X, gamma);
            result.rawSolution_s_gamma = solution_gamma;
            System.out.println("Solution gamma:"+Arrays.toString(solution_gamma));

            ArrayList<ArrayList<Integer>> arrangement = new ArrayList<>();
            for(int i = 0; i < nCar; i++){
                arrangement.add(new ArrayList<>());
            }


            for(Edge2 e : E0){
                for(int l = 0; l < vehicle.size(); l++){
                    if(solution_alpha[e.index][l] > 0.5){
                        //System.out.println("Ride "+e.node2.idx_ride+" is served by "+e.node1.idx_car+"("+l+")" );
                        arrangement.get(vehicle.get(l)).add(e.node2.idx_ride);
                    }
                }
            }
            for(int l = 0; l < vehicle.size(); l++){
                boolean flag = true;
                while (flag){
                    flag = false;
                    if(arrangement.get(vehicle.get(l)).size() < 1)
                        continue;
                    int temp = arrangement.get(vehicle.get(l)).get(arrangement.get(vehicle.get(l)).size()-1);
                    for(Edge2 e : E1){
                        if(solution_alpha[e.index][l] > 0.5 && e.node1.idx_ride == temp){
                            arrangement.get(vehicle.get(l)).add(e.node2.idx_ride);
                            flag = true;
                            System.out.println(vehicle.get(l)+","+temp);
                        }
                    }
                }

            }

            for(int i = 0; i < nCar; i++){
                System.out.println("arrangement for car "+i+": "+arrangement.get(i).toString());
            }


            //System.out.println(Arrays.toString(newArray));
            result.assignment = arrangement;

//            System.out.println("Obj: " + model.get(GRB.DoubleAttr.ObjVal));
            double upperbound = model.get(GRB.DoubleAttr.ObjBound);
            double lowerbound = model.get(GRB.DoubleAttr.ObjVal);
            result.gap_s_phaseII = 1- lowerbound/upperbound;
            // Dispose of model and environment
            result.outputPhaseIIStochastic();
            model.dispose();
            env.dispose();

        } catch (GRBException e) {
            System.out.println("Error code: " + e.getErrorCode() + ". " +
                    e.getMessage());
        }
    }



}
